import copy
import fnmatch
import logging
import time

from salt.utils.vault.exceptions import (
    VaultException,
    VaultInvocationError,
    VaultNotFoundError,
    VaultPermissionDeniedError,
)
from salt.utils.vault.helpers import iso_to_timestamp, timestring_map

log = logging.getLogger(__name__)


class DurationMixin:
    """
    Mixin that handles expiration with time
    """

    def __init__(
        self,
        renewable=False,
        duration=0,
        creation_time=None,
        expire_time=None,
        **kwargs,
    ):
        if "lease_duration" in kwargs:
            duration = kwargs.pop("lease_duration")
        self.renewable = renewable
        self.duration = duration
        creation_time = (
            creation_time if creation_time is not None else round(time.time())
        )
        try:
            creation_time = int(creation_time)
        except ValueError:
            creation_time = iso_to_timestamp(creation_time)
        self.creation_time = creation_time

        expire_time = (
            expire_time if expire_time is not None else round(time.time()) + duration
        )
        try:
            expire_time = int(expire_time)
        except ValueError:
            expire_time = iso_to_timestamp(expire_time)
        self.expire_time = expire_time
        super().__init__(**kwargs)

    def is_renewable(self):
        """
        Checks whether the lease is renewable
        """
        return self.renewable

    def is_valid_for(self, valid_for=0, blur=0):
        """
        Checks whether the entity is valid

        valid_for
            Check whether the entity will still be valid in the future.
            This can be an integer, which will be interpreted as seconds, or a
            time string using the same format as Vault does:
            Suffix ``s`` for seconds, ``m`` for minutes, ``h`` for hours, ``d`` for days.
            Defaults to 0.

        blur
            Allow undercutting ``valid_for`` for this amount of seconds.
            Defaults to 0.
        """
        if not self.duration:
            return True
        delta = self.expire_time - time.time() - timestring_map(valid_for)
        if delta >= 0:
            return True
        return abs(delta) <= blur


class UseCountMixin:
    """
    Mixin that handles expiration with number of uses
    """

    def __init__(self, num_uses=0, use_count=0, **kwargs):
        self.num_uses = num_uses
        self.use_count = use_count
        super().__init__(**kwargs)

    def used(self):
        """
        Increment the use counter by one.
        """
        self.use_count += 1

    def has_uses_left(self, uses=1):
        """
        Check whether this entity has uses left.
        """
        return self.num_uses == 0 or self.num_uses - (self.use_count + uses) >= 0


class DropInitKwargsMixin:
    """
    Mixin that breaks the chain of passing unhandled kwargs up the MRO.
    """

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super().__init__(*args)


class AccessorMixin:
    """
    Mixin that manages accessor information relevant for tokens/secret IDs
    """

    def __init__(self, accessor=None, wrapping_accessor=None, **kwargs):
        # ensure the accessor always points to the actual entity
        if "wrapped_accessor" in kwargs:
            wrapping_accessor = accessor
            accessor = kwargs.pop("wrapped_accessor")
        self.accessor = accessor
        self.wrapping_accessor = wrapping_accessor
        super().__init__(**kwargs)

    def accessor_payload(self):
        if self.accessor is not None:
            return {"accessor": self.accessor}
        raise VaultInvocationError("No accessor information available")


class BaseLease(DurationMixin, DropInitKwargsMixin):
    """
    Base class for leases that expire with time.
    """

    def __init__(self, lease_id, **kwargs):
        self.id = self.lease_id = lease_id
        super().__init__(**kwargs)

    def __str__(self):
        return self.id

    def __repr__(self):
        return repr(self.to_dict())

    def __eq__(self, other):
        try:
            data = other.__dict__
        except AttributeError:
            data = other
        return data == self.__dict__

    def with_renewed(self, **kwargs):
        """
        Partially update the contained data after lease renewal
        """
        attrs = copy.copy(self.__dict__)
        # ensure expire_time is reset properly
        attrs.pop("expire_time")
        attrs.update(kwargs)
        return type(self)(**attrs)

    def to_dict(self):
        """
        Return a dict of all contained attributes
        """
        return self.__dict__


class VaultLease(BaseLease):
    """
    Data object representing a Vault lease.
    """

    def __init__(self, lease_id, data, **kwargs):
        # save lease-associated data
        self.data = data
        super().__init__(lease_id, **kwargs)


class VaultToken(UseCountMixin, AccessorMixin, BaseLease):
    """
    Data object representing an authentication token
    """

    def __init__(self, **kwargs):
        if "client_token" in kwargs:
            # Ensure response data from Vault is accepted as well
            kwargs["lease_id"] = kwargs.pop("client_token")
        super().__init__(**kwargs)

    def is_valid(self, valid_for=0, uses=1):
        """
        Checks whether the token is valid for an amount of time and number of uses

        valid_for
            Check whether the token will still be valid in the future.
            This can be an integer, which will be interpreted as seconds, or a
            time string using the same format as Vault does:
            Suffix ``s`` for seconds, ``m`` for minutes, ``h`` for hours, ``d`` for days.
            Defaults to 0.

        uses
            Check whether the token has at least this number of uses left. Defaults to 1.
        """
        return self.is_valid_for(valid_for) and self.has_uses_left(uses)

    def is_renewable(self):
        """
        Check whether the token is renewable, which requires it
        to be currently valid for at least two uses and renewable
        """
        # Renewing a token deducts a use, hence it does not make sense to
        # renew a token on the last use
        return self.renewable and self.is_valid(uses=2)

    def payload(self):
        """
        Return the payload to use for POST requests using this token
        """
        return {"token": str(self)}

    def serialize_for_minion(self):
        """
        Serialize all necessary data to recreate this object
        into a dict that can be sent to a minion.
        """
        return {
            "client_token": self.id,
            "renewable": self.renewable,
            "lease_duration": self.duration,
            "num_uses": self.num_uses,
            "creation_time": self.creation_time,
            "expire_time": self.expire_time,
        }


class VaultSecretId(UseCountMixin, AccessorMixin, BaseLease):
    """
    Data object representing an AppRole secret ID.
    """

    def __init__(self, **kwargs):
        if "secret_id" in kwargs:
            # Ensure response data from Vault is accepted as well
            kwargs["lease_id"] = kwargs.pop("secret_id")
            kwargs["lease_duration"] = kwargs.pop("secret_id_ttl")
            kwargs["num_uses"] = kwargs.pop("secret_id_num_uses", 0)
            kwargs["accessor"] = kwargs.pop("secret_id_accessor", None)
        if "expiration_time" in kwargs:
            kwargs["expire_time"] = kwargs.pop("expiration_time")
        super().__init__(**kwargs)

    def is_valid(self, valid_for=0, uses=1):
        """
        Checks whether the secret ID is valid for an amount of time and number of uses

        valid_for
            Check whether the secret ID will still be valid in the future.
            This can be an integer, which will be interpreted as seconds, or a
            time string using the same format as Vault does:
            Suffix ``s`` for seconds, ``m`` for minutes, ``h`` for hours, ``d`` for days.
            Defaults to 0.

        uses
            Check whether the secret ID has at least this number of uses left. Defaults to 1.
        """
        return self.is_valid_for(valid_for) and self.has_uses_left(uses)

    def payload(self):
        """
        Return the payload to use for POST requests using this secret ID
        """
        return {"secret_id": str(self)}

    def serialize_for_minion(self):
        """
        Serialize all necessary data to recreate this object
        into a dict that can be sent to a minion.
        """
        return {
            "secret_id": self.id,
            "secret_id_ttl": self.duration,
            "secret_id_num_uses": self.num_uses,
            "creation_time": self.creation_time,
            "expire_time": self.expire_time,
        }


class VaultWrappedResponse(AccessorMixin, BaseLease):
    """
    Data object representing a wrapped response
    """

    def __init__(
        self,
        creation_path,
        **kwargs,
    ):
        if "token" in kwargs:
            # Ensure response data from Vault is accepted as well
            kwargs["lease_id"] = kwargs.pop("token")
            kwargs["lease_duration"] = kwargs.pop("ttl")
        if "renewable" not in kwargs:
            # Not renewable might be incorrect, wrapped tokens are,
            # but we cannot know what was wrapped here.
            kwargs["renewable"] = False
        super().__init__(**kwargs)
        self.creation_path = creation_path

    def serialize_for_minion(self):
        """
        Serialize all necessary data to recreate this object
        into a dict that can be sent to a minion.
        """
        return {
            "wrap_info": {
                "token": self.id,
                "ttl": self.duration,
                "creation_time": self.creation_time,
                "creation_path": self.creation_path,
            },
        }


class LeaseStore:
    """
    Caches leases and handles lease operations
    """

    def __init__(self, client, cache, expire_events=None):
        self.client = client
        self.cache = cache
        self.expire_events = expire_events
        # to update cached leases after renewal/revocation, we need a mapping id => ckey
        self.lease_id_ckey_cache = {}

    def get(
        self,
        ckey,
        valid_for=0,
        renew=True,
        renew_increment=None,
        renew_blur=2,
        revoke=60,
    ):
        """
        Return cached lease or None.

        ckey
            Cache key the lease has been saved in.

        valid_for
            Ensure the returned lease is valid for at least this amount of time.
            This can be an integer, which will be interpreted as seconds, or a
            time string using the same format as Vault does:
            Suffix ``s`` for seconds, ``m`` for minutes, ``h`` for hours, ``d`` for days.
            Defaults to 0.

            .. note::

                This does not take into account token validity, which active leases
                are bound to as well.

        renew
            If the lease is still valid, but not valid for ``valid_for``, attempt to
            renew it. Defaults to true.

        renew_increment
            When renewing, request the lease to be valid for this amount of time from
            the current point of time onwards.
            If unset, will renew the lease by its default validity period and, if
            the renewed lease does not pass ``valid_for``, will try to renew it
            by ``valid_for``.

        renew_blur
            When checking validity after renewal, allow this amount of seconds in leeway
            to account for latency. Especially important when renew_increment is unset
            and the default validity period is less than ``valid_for``.
            Defaults to 2.

        revoke
            If the lease is not valid for ``valid_for`` and renewals
            are disabled or impossible, attempt to have Vault revoke the lease
            after this amount of time and flush the cache. Defaults to 60s.
        """
        if renew_increment is not None and timestring_map(valid_for) > timestring_map(
            renew_increment
        ):
            raise VaultInvocationError(
                "When renew_increment is set, it must be at least valid_for to make sense"
            )

        def check_revoke(lease):
            if self.expire_events is not None:
                self.expire_events(
                    tag=f"vault/lease/{ckey}/expire", data={"valid_for_less": valid_for}
                )
            if revoke:
                self.revoke(lease, delta=revoke)
            return None

        # Since we can renew leases, do not check for future validity in cache
        lease = self.cache.get(ckey, flush=bool(revoke))
        if lease is not None:
            self.lease_id_ckey_cache[str(lease)] = ckey
        if lease is None or lease.is_valid_for(valid_for):
            return lease
        if not renew:
            return check_revoke(lease)
        try:
            lease = self.renew(lease, increment=renew_increment, raise_all_errors=False)
        except VaultNotFoundError:
            # The cached lease was already revoked
            return check_revoke(lease)
        if not lease.is_valid_for(valid_for, blur=renew_blur):
            if renew_increment is not None:
                # valid_for cannot possibly be respected
                return check_revoke(lease)
            # Maybe valid_for is greater than the default validity period, so check if
            # the lease can be renewed by valid_for
            try:
                lease = self.renew(lease, increment=valid_for, raise_all_errors=False)
            except VaultNotFoundError:
                # The cached lease was already revoked
                return check_revoke(lease)
            if not lease.is_valid_for(valid_for, blur=renew_blur):
                return check_revoke(lease)
        return lease

    def list(self):
        """
        List all cached leases.
        """
        return self.cache.list()

    def lookup(self, lease):
        """
        Lookup lease meta information.

        lease
            A lease ID or VaultLease object to look up.
        """
        endpoint = "sys/leases/lookup"
        payload = {"lease_id": str(lease)}
        return self.client.post(endpoint, payload=payload)

    def renew(self, lease, increment=None, raise_all_errors=True, _store=True):
        """
        Renew a lease.

        lease
            A lease ID or VaultLease object to renew.

        increment
            Request the lease to be valid for this amount of time from the current
            point of time onwards. Can also be used to reduce the validity period.
            The server might not honor this increment.
            Can be an integer (seconds) or a time string like ``1h``. Optional.

        raise_all_errors
            When ``lease`` is a VaultLease and the renewal does not succeed,
            do not catch exceptions. If this is false, the lease will be returned
            unmodified if the exception does not indicate it is invalid (NotFound).
            Defaults to true.
        """
        endpoint = "sys/leases/renew"
        payload = {"lease_id": str(lease)}
        if increment is not None:
            payload["increment"] = int(timestring_map(increment))
        if not isinstance(lease, VaultLease) and lease in self.lease_id_ckey_cache:
            lease = self.cache.get(self.lease_id_ckey_cache[lease], flush=False)
            if lease is None:
                raise VaultNotFoundError("Lease is already expired")
        try:
            ret = self.client.post(endpoint, payload=payload)
        except VaultException as err:
            if raise_all_errors or not isinstance(lease, VaultLease):
                raise
            if isinstance(err, VaultNotFoundError):
                raise
            return lease

        if _store and isinstance(lease, VaultLease):
            # Do not overwrite data of renewed leases!
            ret.pop("data", None)
            new_lease = lease.with_renewed(**ret)
            if str(new_lease) in self.lease_id_ckey_cache:
                self.store(self.lease_id_ckey_cache[str(new_lease)], new_lease)
            return new_lease
        return ret

    def renew_cached(self, match, increment=None):
        """
        Renew cached leases.

        match
            Only renew cached leases whose ckey matches this glob pattern.
            Defaults to ``*``.

        increment
            Request the leases to be valid for this amount of time from the current
            point of time onwards. Can also be used to reduce the validity period.
            The server might not honor this increment.
            Can be an integer (seconds) or a time string like ``1h``. Optional.
        """
        failed = []
        for ckey in self.list():
            if not fnmatch.fnmatch(ckey, match):
                continue
            lease = self.cache.get(ckey, flush=True)
            if lease is None:
                continue
            self.lease_id_ckey_cache[str(lease)] = ckey
            try:
                self.renew(lease, increment=increment)
            except (VaultPermissionDeniedError, VaultNotFoundError) as err:
                log.warning("Failed renewing cached lease: %s", type(err).__name__)
                log.debug("Lease ID was: %s", lease)
                failed.append(ckey)
        if failed:
            raise VaultException(f"Failed renewing some leases: {list(failed)}")
        return True

    def revoke(self, lease, delta=60):
        """
        Revoke a lease. Will also remove the cached lease,
        if it has been requested from this LeaseStore before.

        lease
            A lease ID or VaultLease object to revoke.

        delta
            Time after which the lease should be requested
            to be revoked by Vault.
            Defaults to 60s.
        """
        try:
            # 0 would attempt a complete renewal
            self.renew(lease, increment=delta or 1, _store=False)
        except VaultNotFoundError:
            pass

        if str(lease) in self.lease_id_ckey_cache:
            self.cache.flush(self.lease_id_ckey_cache.pop(str(lease)))
        return True

    def revoke_cached(
        self,
        match="*",
        delta=60,
        flush_on_failure=True,
    ):
        """
        Revoke cached leases.

        match
            Only revoke cached leases whose ckey matches this glob pattern.
            Defaults to ``*``.

        delta
            Time after which the leases should be revoked by Vault.
            Defaults to 60s.

        flush_on_failure
            If a revocation fails, remove the lease from cache anyways.
            Defaults to true.
        """
        failed = []
        for ckey in self.list():
            if not fnmatch.fnmatch(ckey, match):
                continue
            lease = self.cache.get(ckey, flush=True)
            if lease is None:
                continue
            self.lease_id_ckey_cache[str(lease)] = ckey
            try:
                self.revoke(lease, delta=delta)
            except VaultPermissionDeniedError:
                failed.append(ckey)
                if flush_on_failure:
                    # Forget the lease and let Vault's automatic revocation handle it
                    self.cache.flush(self.lease_id_ckey_cache.pop(str(lease)))
        if failed:
            raise VaultException(f"Failed revoking some leases: {list(failed)}")
        return True

    def store(self, ckey, lease):
        """
        Cache a lease.

        ckey
            The cache key the lease should be saved in.

        lease
            A lease ID or VaultLease object to store.
        """
        self.cache.store(ckey, lease)
        self.lease_id_ckey_cache[str(lease)] = ckey
        return True



import contextlib
import io
import stat

import pytest

import salt.config
import salt.daemons.masterapi as masterapi
import salt.utils.platform
from tests.support.mock import MagicMock, patch


def gen_permissions(owner="", group="", others=""):
    """
    Helper method to generate file permission bits
    Usage: gen_permissions('rw', 'r', 'r')
    """
    ret = 0
    for c in owner:
        ret |= getattr(stat, f"S_I{c.upper()}USR", 0)
    for c in group:
        ret |= getattr(stat, f"S_I{c.upper()}GRP", 0)
    for c in others:
        ret |= getattr(stat, f"S_I{c.upper()}OTH", 0)
    return ret


@contextlib.contextmanager
def patch_check_permissions(
    auto_key, stats, uid=1, groups=None, is_windows=False, permissive_pki=False
):
    def os_stat_mock(filename):
        fmode = MagicMock()
        fstats = stats.get(filename, {})
        fmode.st_mode = fstats.get("mode", 0)
        fmode.st_gid = fstats.get("gid", 0)
        return fmode

    if not groups:
        groups = [uid]

    auto_key.opts["permissive_pki_access"] = permissive_pki
    if salt.utils.platform.is_windows():
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=True)):
            yield
    else:
        with patch("os.stat", os_stat_mock), patch(
            "os.getuid", MagicMock(return_value=uid)
        ), patch("salt.utils.user.get_gid_list", MagicMock(return_value=groups)), patch(
            "salt.utils.platform.is_windows", MagicMock(return_value=is_windows)
        ):
            yield


@pytest.fixture
def auto_key():
    opts = salt.config.master_config(None)
    opts["user"] = "test_user"
    return masterapi.AutoKey(opts)


@pytest.fixture
def local_funcs():
    opts = salt.config.master_config(None)
    return masterapi.LocalFuncs(opts, "test-key")


def test_check_permissions_windows(auto_key):
    """
    Assert that all files are accepted on windows
    """
    stats = {
        "testfile": {
            "mode": gen_permissions("rwx", "rwx", "rwx"),
            "gid": 2,
        },
    }
    with patch_check_permissions(auto_key, stats, uid=0, is_windows=True):
        assert auto_key.check_permissions("testfile") is True


def test_check_permissions_others_can_write(auto_key):
    """
    Assert that no file is accepted, when others can write to it
    """
    stats = {
        "testfile": {
            "mode": gen_permissions("", "", "w"),
            "gid": 1,
        },
    }
    with patch_check_permissions(auto_key, stats, permissive_pki=True):
        if salt.utils.platform.is_windows():
            assert auto_key.check_permissions("testfile") is True
        else:
            assert auto_key.check_permissions("testfile") is False


def test_check_permissions_group_can_write_not_permissive(auto_key):
    """
    Assert that a file is accepted, when group can write to it and
    permissive_pki_access=False
    """
    stats = {
        "testfile": {
            "mode": gen_permissions("w", "w", ""),
            "gid": 1,
        },
    }
    with patch_check_permissions(auto_key, stats):
        if salt.utils.platform.is_windows():
            assert auto_key.check_permissions("testfile") is True
        else:
            assert auto_key.check_permissions("testfile") is False


def test_check_permissions_group_can_write_permissive(auto_key):
    """
    Assert that a file is accepted, when group can write to it and
    permissive_pki_access=True
    """
    stats = {
        "testfile": {
            "mode": gen_permissions("w", "w", ""),
            "gid": 1,
        },
    }

    with patch_check_permissions(auto_key, stats, permissive_pki=True):
        assert auto_key.check_permissions("testfile") is True


def test_check_permissions_group_can_write_permissive_root_in_group(auto_key):
    """
    Assert that a file is accepted, when group can write to it,
    permissive_pki_access=False, salt is root and in the file owning group
    """
    stats = {
        "testfile": {
            "mode": gen_permissions("w", "w", ""),
            "gid": 0,
        },
    }
    with patch_check_permissions(auto_key, stats, uid=0, permissive_pki=True):
        assert auto_key.check_permissions("testfile") is True


def test_check_permissions_group_can_write_permissive_root_not_in_group(auto_key):
    """
    Assert that no file is accepted, when group can write to it,
    permissive_pki_access=False, salt is root and **not** in the file owning
    group
    """
    stats = {
        "testfile": {
            "mode": gen_permissions("w", "w", ""),
            "gid": 1,
        },
    }
    with patch_check_permissions(auto_key, stats, uid=0, permissive_pki=True):
        if salt.utils.platform.is_windows():
            assert auto_key.check_permissions("testfile") is True
        else:
            assert auto_key.check_permissions("testfile") is False


def test_check_permissions_only_owner_can_write(auto_key):
    """
    Assert that a file is accepted, when only the owner can write to it
    """
    stats = {
        "testfile": {
            "mode": gen_permissions("w", "", ""),
            "gid": 1,
        },
    }
    with patch_check_permissions(auto_key, stats):
        assert auto_key.check_permissions("testfile") is True


def test_check_permissions_only_owner_can_write_root(auto_key):
    """
    Assert that a file is accepted, when only the owner can write to it and salt is root
    """
    stats = {
        "testfile": {
            "mode": gen_permissions("w", "", ""),
            "gid": 0,
        },
    }
    with patch_check_permissions(auto_key, stats, uid=0):
        assert auto_key.check_permissions("testfile") is True


def _test_check_autosign_grains(
    test_func,
    auto_key,
    file_content="test_value",
    file_name="test_grain",
    autosign_grains_dir="test_dir",
    permissions_ret=True,
):
    """
    Helper function for testing autosign_grains().

    Patches ``os.walk`` to return only ``file_name`` and ``salt.utils.files.fopen`` to open a
    mock file with ``file_content`` as content. Optionally sets ``opts`` values.
    Then executes test_func. The ``os.walk`` and ``salt.utils.files.fopen`` mock objects
    are passed to the function as arguments.
    """
    if autosign_grains_dir:
        auto_key.opts["autosign_grains_dir"] = autosign_grains_dir
    mock_file = io.StringIO(file_content)
    mock_dirs = [(None, None, [file_name])]

    with patch("os.walk", MagicMock(return_value=mock_dirs)) as mock_walk, patch(
        "salt.utils.files.fopen", MagicMock(return_value=mock_file)
    ) as mock_open, patch(
        "salt.daemons.masterapi.AutoKey.check_permissions",
        MagicMock(return_value=permissions_ret),
    ) as mock_permissions:
        test_func(mock_walk, mock_open, mock_permissions)


def test_check_autosign_grains_no_grains(auto_key):
    """
    Asserts that autosigning from grains fails when no grain values are passed.
    """

    def test_func(mock_walk, mock_open, mock_permissions):
        assert auto_key.check_autosign_grains(None) is False
        assert mock_walk.call_count == 0
        assert mock_open.call_count == 0
        assert mock_permissions.call_count == 0

        assert auto_key.check_autosign_grains({}) is False
        assert mock_walk.call_count == 0
        assert mock_open.call_count == 0
        assert mock_permissions.call_count == 0

    _test_check_autosign_grains(test_func, auto_key)


def test_check_autosign_grains_no_autosign_grains_dir(auto_key):
    """
    Asserts that autosigning from grains fails when the \'autosign_grains_dir\' config option
    is undefined.
    """

    def test_func(mock_walk, mock_open, mock_permissions):
        assert auto_key.check_autosign_grains({"test_grain": "test_value"}) is False
        assert mock_walk.call_count == 0
        assert mock_open.call_count == 0
        assert mock_permissions.call_count == 0

    _test_check_autosign_grains(test_func, auto_key, autosign_grains_dir=None)


@pytest.mark.parametrize("grain_value", ["test_value", 123, True])
def test_check_autosign_grains_accept(grain_value, auto_key):
    """
    Asserts that autosigning from grains passes when a matching grain value is in an
    autosign_grain file.
    """

    def test_func(*args):
        assert auto_key.check_autosign_grains({"test_grain": grain_value}) is True

    file_content = f"#test_ignore\n{grain_value}"
    _test_check_autosign_grains(test_func, auto_key, file_content=file_content)


def test_check_autosign_grains_accept_not(auto_key):
    """
    Asserts that autosigning from grains fails when the grain value is not in the
    autosign_grain files.
    """

    def test_func(*args):
        assert auto_key.check_autosign_grains({"test_grain": "test_invalid"}) is False

    file_content = "#test_invalid\ntest_value"
    _test_check_autosign_grains(test_func, auto_key, file_content=file_content)


def test_check_autosign_grains_invalid_file_permissions(auto_key):
    """
    Asserts that autosigning from grains fails when the grain file has the wrong permissions.
    """

    def test_func(*args):
        assert auto_key.check_autosign_grains({"test_grain": "test_value"}) is False

    file_content = "#test_ignore\ntest_value"
    _test_check_autosign_grains(
        test_func, auto_key, file_content=file_content, permissions_ret=False
    )



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import pytest

import salt.states.rabbitmq_cluster as rabbitmq_cluster
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {rabbitmq_cluster: {}}


def test_joined():
    """
    Test to ensure the current node joined
    to a cluster with node user@host
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(side_effect=[["rahulha@salt"], [""], [""]])
    with patch.dict(rabbitmq_cluster.__salt__, {"rabbitmq.cluster_status": mock}):
        ret.update({"comment": "Already in cluster"})
        assert rabbitmq_cluster.joined("salt", "salt", "rahulha") == ret

        with patch.dict(rabbitmq_cluster.__opts__, {"test": True}):
            ret.update(
                {
                    "result": None,
                    "comment": "Node is set to join cluster rahulha@salt",
                    "changes": {"new": "rahulha@salt", "old": ""},
                }
            )
            assert rabbitmq_cluster.joined("salt", "salt", "rahulha") == ret

        with patch.dict(rabbitmq_cluster.__opts__, {"test": False}):
            mock = MagicMock(return_value={"Error": "ERR"})
            with patch.dict(rabbitmq_cluster.__salt__, {"rabbitmq.join_cluster": mock}):
                ret.update({"result": False, "comment": "ERR", "changes": {}})
                assert rabbitmq_cluster.joined("salt", "salt", "rahulha") == ret



"""
This is the default pillar exact matcher for compound matches.

There is no minion-side equivalent for this, so consequently there is no ``match()``
function below, only an ``mmatch()``
"""

import logging

import salt.utils.minions

log = logging.getLogger(__name__)


def mmatch(expr, delimiter, greedy, opts=None):
    """
    Return the minions found by looking via pillar
    """
    if not opts:
        opts = __opts__

    ckminions = salt.utils.minions.CkMinions(opts)
    return ckminions._check_compound_minions(expr, delimiter, greedy, pillar_exact=True)



import os
import shutil
import subprocess

import pytest

import salt.utils.platform
from tests.conftest import FIPS_TESTRUN


@pytest.fixture
def salt_master_1(request, salt_factories):
    config_defaults = {
        "open_mode": True,
        "transport": request.config.getoption("--transport"),
    }
    config_overrides = {
        "interface": "127.0.0.1",
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }

    factory = salt_factories.salt_master_daemon(
        "master-1",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture
def salt_master_2(salt_factories, salt_master_1):
    if salt.utils.platform.is_darwin() or salt.utils.platform.is_freebsd():
        subprocess.check_output(["ifconfig", "lo0", "alias", "127.0.0.2", "up"])

    config_defaults = {
        "open_mode": True,
        "transport": salt_master_1.config["transport"],
    }
    config_overrides = {
        "interface": "127.0.0.2",
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }

    # Use the same ports for both masters, they are binding to different interfaces
    for key in (
        "ret_port",
        "publish_port",
    ):
        config_overrides[key] = salt_master_1.config[key]
    factory = salt_factories.salt_master_daemon(
        "master-2",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )

    # The secondary salt master depends on the primarily salt master fixture
    # because we need to clone the keys
    for keyfile in ("master.pem", "master.pub"):
        shutil.copyfile(
            os.path.join(salt_master_1.config["pki_dir"], keyfile),
            os.path.join(factory.config["pki_dir"], keyfile),
        )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture
def salt_minion_1(salt_master_1, salt_master_2):
    config_defaults = {
        "transport": salt_master_1.config["transport"],
    }

    master_1_port = salt_master_1.config["ret_port"]
    master_1_addr = salt_master_1.config["interface"]
    master_2_port = salt_master_2.config["ret_port"]
    master_2_addr = salt_master_2.config["interface"]
    config_overrides = {
        "master": [
            f"{master_1_addr}:{master_1_port}",
            f"{master_2_addr}:{master_2_port}",
        ],
        "test.foo": "baz",
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    factory = salt_master_1.salt_minion_daemon(
        "minion-1",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory



import pytest

import salt.engines
from tests.support.mock import MagicMock, patch


@pytest.fixture
def kwargs():
    opts = {"__role": "minion"}
    name = "foobar"
    fun = f"{name}.start"
    config = funcs = runners = proxy = {}
    return dict(
        opts=opts,
        name=name,
        fun=fun,
        config=config,
        funcs=funcs,
        runners=runners,
        proxy=proxy,
    )


def test_engine_module_name(kwargs):
    engine = salt.engines.Engine(**kwargs)
    assert engine.name == kwargs["name"]


def test_engine_title_set(kwargs):
    engine = salt.engines.Engine(**kwargs)
    with patch("salt.utils.process.appendproctitle", MagicMock()) as mm:
        engine.run()
    mm.assert_called_with(kwargs["name"])



"""
Tests for salt.modules.zfs on Solaris
"""

import pytest

import salt.loader
import salt.modules.zfs as zfs
import salt.utils.zfs
from tests.support.mock import MagicMock, patch
from tests.support.zfs import ZFSMockData


@pytest.fixture
def utils_patch():
    return ZFSMockData().get_patched_utils()


@pytest.fixture
def configure_loader_modules(minion_opts):
    utils = salt.loader.utils(minion_opts, whitelist=["zfs"])
    zfs_obj = {
        zfs: {
            "__opts__": minion_opts,
            "__grains__": {
                "osarch": "sparcv9",
                "os_family": "Solaris",
                "osmajorrelease": 11,
                "kernel": "SunOS",
                "kernelrelease": 5.11,
            },
            "__utils__": utils,
        }
    }

    return zfs_obj


@pytest.mark.skip_unless_on_sunos
def test_get_success_solaris():
    """
    Tests zfs get success
    """

    cmd_out = {
        "pid": 7278,
        "retcode": 0,
        "stdout": "testpool\tmountpoint\t/testpool\tdefault",
        "stderr": "",
    }

    run_all_mock = MagicMock(return_value=cmd_out)
    patches = {
        "cmd.run_all": run_all_mock,
    }
    with patch.dict(zfs.__salt__, patches):
        with patch("sys.platform", MagicMock(return_value="sunos5")):
            result = zfs.get("testpool", type="filesystem", properties="mountpoint")
            assert result == {
                "testpool": {
                    "mountpoint": {"value": "/testpool", "source": "default"},
                },
            }
    run_all_mock.assert_called_once_with(
        "/usr/sbin/zfs get -H -o name,property,value,source mountpoint testpool",
        python_shell=False,
    )



import io

import pytest

import salt.state
from salt import template
from tests.support.mock import MagicMock, patch


@pytest.fixture
def render_dict():
    return {
        "jinja": "fake_jinja_func",
        "json": "fake_json_func",
        "mako": "fake_make_func",
    }


def test_compile_template_str_mkstemp_cleanup(minion_opts):
    minion_opts["file_client"] = "local"
    with patch("os.unlink", MagicMock()) as unlinked:
        _state = salt.state.State(minion_opts)
        ret = template.compile_template_str(
            "{'val':'test'}",
            _state.rend,
            _state.opts["renderer"],
            _state.opts["renderer_blacklist"],
            _state.opts["renderer_whitelist"],
        )
        assert ret == {"val": "test"}
        unlinked.assert_called_once()


def test_compile_template_bad_type():
    """
    Test to ensure that unsupported types cannot be passed to the template compiler
    """
    ret = template.compile_template(["1", "2", "3"], None, None, None, None)
    assert ret == {}


def test_compile_template_preserves_windows_newlines():
    """
    Test to ensure that a file with Windows newlines, when rendered by a
    template renderer, does not eat the CR character.
    """

    def _get_rend(renderer, value):
        """
        We need a new MagicMock each time since we're dealing with StringIO
        objects which are read like files.
        """
        return {renderer: MagicMock(return_value=io.StringIO(value))}

    input_data_windows = "foo\r\nbar\r\nbaz\r\n"
    input_data_non_windows = input_data_windows.replace("\r\n", "\n")
    renderer = "test"
    blacklist = whitelist = []

    ret = template.compile_template(
        ":string:",
        _get_rend(renderer, input_data_non_windows),
        renderer,
        blacklist,
        whitelist,
        input_data=input_data_windows,
    ).read()
    # Even though the mocked renderer returned a string without the windows
    # newlines, the compiled template should still have them.
    assert ret == input_data_windows

    # Now test that we aren't adding them in unnecessarily.
    ret = template.compile_template(
        ":string:",
        _get_rend(renderer, input_data_non_windows),
        renderer,
        blacklist,
        whitelist,
        input_data=input_data_non_windows,
    ).read()
    assert ret == input_data_non_windows

    # Finally, ensure that we're not unnecessarily replacing the \n with
    # \r\n in the event that the renderer returned a string with the
    # windows newlines intact.
    ret = template.compile_template(
        ":string:",
        _get_rend(renderer, input_data_windows),
        renderer,
        blacklist,
        whitelist,
        input_data=input_data_windows,
    ).read()
    assert ret == input_data_windows


def test_check_render_pipe_str(render_dict):
    """
    Check that all renderers specified in the pipe string are available.
    """
    ret = template.check_render_pipe_str("jinja|json", render_dict, None, None)
    assert ("fake_jinja_func", "") in ret
    assert ("fake_json_func", "") in ret
    assert ("OBVIOUSLY_NOT_HERE", "") not in ret


def test_check_renderer_blacklisting(render_dict):
    """
    Check that all renderers specified in the pipe string are available.
    """
    ret = template.check_render_pipe_str("jinja|json", render_dict, ["jinja"], None)
    assert ret == [("fake_json_func", "")]
    ret = template.check_render_pipe_str("jinja|json", render_dict, None, ["jinja"])
    assert ret == [("fake_jinja_func", "")]
    ret = template.check_render_pipe_str(
        "jinja|json", render_dict, ["jinja"], ["jinja"]
    )
    assert ret == []
    ret = template.check_render_pipe_str(
        "jinja|json", render_dict, ["jinja"], ["jinja", "json"]
    )
    assert ret == [("fake_json_func", "")]



"""
tests.unit.context_test
~~~~~~~~~~~~~~~~~~~~~~~
"""

import salt.utils.json
from salt.utils.context import NamespacedDictWrapper
from tests.support.unit import TestCase


class NamespacedDictWrapperTests(TestCase):
    PREFIX = "prefix"

    def setUp(self):
        self._dict = {}

    def test_single_key(self):
        self._dict["prefix"] = {"foo": "bar"}
        w = NamespacedDictWrapper(self._dict, "prefix")
        self.assertEqual(w["foo"], "bar")

    def test_multiple_key(self):
        self._dict["prefix"] = {"foo": {"bar": "baz"}}
        w = NamespacedDictWrapper(self._dict, ("prefix", "foo"))
        self.assertEqual(w["bar"], "baz")

    def test_json_dumps_single_key(self):
        self._dict["prefix"] = {"foo": {"bar": "baz"}}
        w = NamespacedDictWrapper(self._dict, "prefix")
        self.assertEqual(salt.utils.json.dumps(w), '{"foo": {"bar": "baz"}}')

    def test_json_dumps_multiple_key(self):
        self._dict["prefix"] = {"foo": {"bar": "baz"}}
        w = NamespacedDictWrapper(self._dict, ("prefix", "foo"))
        self.assertEqual(salt.utils.json.dumps(w), '{"bar": "baz"}')



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import textwrap

import pytest

import salt.config
import salt.loader
import salt.states.boto_sqs as boto_sqs
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    utils = salt.loader.utils(
        minion_opts,
        whitelist=["boto3", "yaml", "args", "systemd", "path", "platform"],
        context={},
    )
    return {boto_sqs: {"__utils__": utils}}


def test_present():
    """
    Test to ensure the SQS queue exists.
    """
    name = "mysqs"
    attributes = {"DelaySeconds": 20}
    base_ret = {"name": name, "changes": {}}

    mock = MagicMock(
        side_effect=[{"result": b} for b in [False, False, True, True]],
    )
    mock_bool = MagicMock(return_value={"error": "create error"})
    mock_attr = MagicMock(return_value={"result": {}})
    with patch.dict(
        boto_sqs.__salt__,
        {
            "boto_sqs.exists": mock,
            "boto_sqs.create": mock_bool,
            "boto_sqs.get_attributes": mock_attr,
        },
    ):
        with patch.dict(boto_sqs.__opts__, {"test": False}):
            comt = [
                "Failed to create SQS queue {}: create error".format(
                    name,
                )
            ]
            ret = base_ret.copy()
            ret.update({"result": False, "comment": comt})
            assert boto_sqs.present(name) == ret

        with patch.dict(boto_sqs.__opts__, {"test": True}):
            comt = [f"SQS queue {name} is set to be created."]
            ret = base_ret.copy()
            ret.update(
                {
                    "result": None,
                    "comment": comt,
                    "changes": {"old": None, "new": "mysqs"},
                }
            )
            assert boto_sqs.present(name) == ret
            diff = textwrap.dedent(
                """\
                ---
                +++
                @@ -1 +1 @@
                -{}
                +DelaySeconds: 20

            """
            ).splitlines()
            for idx in (0, 1):
                diff[idx] += " "
            diff = "\n".join(diff)

            comt = [
                "SQS queue mysqs present.",
                "Attribute(s) DelaySeconds set to be updated:\n{}".format(
                    diff,
                ),
            ]
            ret.update({"comment": comt, "changes": {"attributes": {"diff": diff}}})
            assert boto_sqs.present(name, attributes) == ret

        comt = ["SQS queue mysqs present."]
        ret = base_ret.copy()
        ret.update({"result": True, "comment": comt})
        assert boto_sqs.present(name) == ret


def test_absent():
    """
    Test to ensure the named sqs queue is deleted.
    """
    name = "test.example.com."
    base_ret = {"name": name, "changes": {}}

    mock = MagicMock(side_effect=[{"result": False}, {"result": True}])
    with patch.dict(boto_sqs.__salt__, {"boto_sqs.exists": mock}):
        comt = f"SQS queue {name} does not exist in None."
        ret = base_ret.copy()
        ret.update({"result": True, "comment": comt})
        assert boto_sqs.absent(name) == ret

        with patch.dict(boto_sqs.__opts__, {"test": True}):
            comt = f"SQS queue {name} is set to be removed."
            ret = base_ret.copy()
            ret.update(
                {
                    "result": None,
                    "comment": comt,
                    "changes": {"old": name, "new": None},
                }
            )
            assert boto_sqs.absent(name) == ret



import logging

import pytest

import salt.pillar.etcd_pillar as etcd_pillar
from salt.utils.etcd_util import get_conn
from tests.support.pytest.etcd import *  # pylint: disable=wildcard-import,unused-wildcard-import

docker = pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skip_if_binaries_missing("docker", "dockerd", check_all=False),
]


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        etcd_pillar: {
            "__opts__": minion_opts,
        },
    }


@pytest.fixture(scope="module")
def minion_config_overrides(etcd_profile):
    return etcd_profile


@pytest.fixture(scope="module")
def etcd_client(minion_opts, profile_name):
    return get_conn(minion_opts, profile=profile_name)


@pytest.fixture(scope="module")
def prefix():
    return "/salt/pillar/test"


@pytest.fixture(autouse=True)
def cleanup_prefixed_entries(etcd_client, prefix):
    """
    Cleanup after each test to ensure a consistent starting state.
    """
    try:
        assert etcd_client.get(prefix, recurse=True) is None
        yield
    finally:
        etcd_client.delete(prefix, recurse=True)


def test_ext_pillar(subtests, profile_name, prefix, etcd_client):
    """
    Test ext_pillar functionality
    """
    updated = {
        "1": "not one",
        "2": {
            "3": "two-three",
            "4": "two-four",
        },
    }
    etcd_client.update(updated, path=prefix)

    with subtests.test("We should be able to use etcd as an external pillar"):
        expected = {
            "salt": {
                "pillar": {
                    "test": updated,
                },
            },
        }
        assert etcd_pillar.ext_pillar("minion_id", {}, profile_name) == expected



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import pytest

import salt.states.win_dns_client as win_dns_client
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_dns_client: {}}


def test_dns_exists():
    """
    Test to configure the DNS server list in the specified interface
    """
    ret = {"name": "salt", "changes": {}, "result": False, "comment": ""}
    with patch.dict(win_dns_client.__opts__, {"test": False}):
        ret.update(
            {
                "changes": {
                    "Servers Added": [],
                    "Servers Removed": [],
                    "Servers Reordered": [],
                },
                "comment": "servers entry is not a list !",
            }
        )
        assert win_dns_client.dns_exists("salt") == ret

        mock = MagicMock(return_value=[2, "salt"])
        with patch.dict(
            win_dns_client.__salt__, {"win_dns_client.get_dns_servers": mock}
        ):
            ret.update(
                {
                    "changes": {},
                    "comment": repr([2, "salt"]) + " are already configured",
                    "result": True,
                }
            )
            assert win_dns_client.dns_exists("salt", [2, "salt"]) == ret

            mock = MagicMock(side_effect=[False, True, True])
            with patch.dict(win_dns_client.__salt__, {"win_dns_client.add_dns": mock}):
                ret.update(
                    {
                        "comment": "Failed to add 1 as DNS server number 1",
                        "result": False,
                    }
                )
                assert win_dns_client.dns_exists("salt", [1, "salt"]) == ret

                mock = MagicMock(return_value=False)
                with patch.dict(
                    win_dns_client.__salt__, {"win_dns_client.rm_dns": mock}
                ):
                    ret.update(
                        {
                            "changes": {
                                "Servers Added": ["a"],
                                "Servers Removed": [],
                                "Servers Reordered": [],
                            },
                            "comment": "Failed to remove 2 from DNS server list",
                        }
                    )
                    assert win_dns_client.dns_exists("salt", ["a"], "a", 1) == ret

                ret.update({"comment": "DNS Servers have been updated", "result": True})
                assert win_dns_client.dns_exists("salt", ["a"]) == ret


def test_dns_dhcp():
    """
    Test to configure the DNS server list from DHCP Server
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
    mock = MagicMock(side_effect=["dhcp", "salt", "salt"])
    with patch.dict(win_dns_client.__salt__, {"win_dns_client.get_dns_config": mock}):
        ret.update(
            {"comment": "Local Area Connection already configured with DNS from DHCP"}
        )
        assert win_dns_client.dns_dhcp("salt") == ret

        with patch.dict(win_dns_client.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "",
                    "result": None,
                    "changes": {"dns": "configured from DHCP"},
                }
            )
            assert win_dns_client.dns_dhcp("salt") == ret

        with patch.dict(win_dns_client.__opts__, {"test": False}):
            mock = MagicMock(return_value=True)
            with patch.dict(win_dns_client.__salt__, {"win_dns_client.dns_dhcp": mock}):
                ret.update({"result": True})
                assert win_dns_client.dns_dhcp("salt") == ret


def test_primary_suffix():
    """
    Test to configure the global primary DNS suffix of a DHCP client.
    """
    ret = {"name": "salt", "changes": {}, "result": False, "comment": ""}
    ret.update({"comment": "'updates' must be a boolean value"})
    assert win_dns_client.primary_suffix("salt", updates="a") == ret

    mock = MagicMock(
        side_effect=[
            {"vdata": "a"},
            {"vdata": False},
            {"vdata": "b"},
            {"vdata": False},
        ]
    )
    with patch.dict(win_dns_client.__utils__, {"reg.read_value": mock}):
        ret.update({"comment": "No changes needed", "result": True})
        assert win_dns_client.primary_suffix("salt", "a") == ret

        mock = MagicMock(return_value=True)
        with patch.dict(win_dns_client.__utils__, {"reg.set_value": mock}):
            ret.update(
                {
                    "changes": {"new": {"suffix": "a"}, "old": {"suffix": "b"}},
                    "comment": "Updated primary DNS suffix (a)",
                }
            )
            assert win_dns_client.primary_suffix("salt", "a") == ret



import pytest

import salt.pillar.nodegroups as nodegroups
from tests.support.mock import MagicMock, patch


@pytest.fixture
def fake_minion_id():
    return "fake_id"


@pytest.fixture
def fake_nodegroups(fake_minion_id):
    return {
        "groupA": fake_minion_id,
        "groupB": "another_minion_id",
    }


@pytest.fixture
def fake_pillar_name():
    return "fake_pillar_name"


@pytest.fixture
def configure_loader_modules(fake_minion_id, fake_nodegroups):
    fake_opts = {
        "cache": "localfs",
        "nodegroups": fake_nodegroups,
        "id": fake_minion_id,
    }
    return {nodegroups: {"__opts__": fake_opts}}


def _runner(expected_ret, fake_minion_id, fake_pillar_name, pillar_name=None):
    def _side_effect(group_sel, t):
        if group_sel.find(fake_minion_id) != -1:
            return {"minions": [fake_minion_id], "missing": []}
        return {"minions": ["another_minion_id"], "missing": []}

    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(side_effect=_side_effect),
    ):
        pillar_name = pillar_name or fake_pillar_name
        actual_ret = nodegroups.ext_pillar(fake_minion_id, {}, pillar_name=pillar_name)
        assert actual_ret == expected_ret


def test_succeeds(fake_pillar_name, fake_minion_id):
    ret = {fake_pillar_name: ["groupA"]}
    _runner(ret, fake_minion_id, fake_pillar_name)



"""
Test the ssh module
"""

import os
import shutil

import pytest
import requests

import salt.utils.files
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.runtests import RUNTIME_VARS

GITHUB_FINGERPRINT = "b8:d8:95:ce:d9:2c:0a:c0:e1:71:cd:2e:f5:ef:01:ba:34:17:55:4a:4a:64:80:d3:31:cc:c2:be:3d:ed:0f:6b"


def check_status():
    """
    Check the status of Github for remote operations
    """
    try:
        return requests.get("https://github.com", timeout=60).status_code == 200
    except Exception:  # pylint: disable=broad-except
        return False


# @pytest.mark.windows_whitelisted
# De-whitelist windows since it's hanging on the newer windows golden images
@pytest.mark.skip_if_binaries_missing("ssh", "ssh-keygen", check_all=True)
class SSHModuleTest(ModuleCase):
    """
    Test the ssh module
    """

    @classmethod
    def setUpClass(cls):
        cls.subsalt_dir = os.path.join(RUNTIME_VARS.TMP, "subsalt")
        cls.authorized_keys = os.path.join(cls.subsalt_dir, "authorized_keys")
        cls.known_hosts = os.path.join(cls.subsalt_dir, "known_hosts")

    def setUp(self):
        """
        Set up the ssh module tests
        """
        if not check_status():
            self.skipTest("External source, github.com is down")
        super().setUp()
        if not os.path.isdir(self.subsalt_dir):
            os.makedirs(self.subsalt_dir)

        known_hosts_file = os.path.join(RUNTIME_VARS.FILES, "ssh", "known_hosts")
        with salt.utils.files.fopen(known_hosts_file) as fd:
            self.key = fd.read().strip().splitlines()[0].split()[-1]

    def tearDown(self):
        """
        Tear down the ssh module tests
        """
        if os.path.isdir(self.subsalt_dir):
            shutil.rmtree(self.subsalt_dir)
        super().tearDown()
        del self.key

    @pytest.mark.slow_test
    def test_auth_keys(self):
        """
        test ssh.auth_keys
        """
        shutil.copyfile(
            os.path.join(RUNTIME_VARS.FILES, "ssh", "authorized_keys"),
            self.authorized_keys,
        )
        user = "root"
        if salt.utils.platform.is_windows():
            user = "Administrator"
        ret = self.run_function("ssh.auth_keys", [user, self.authorized_keys])
        self.assertEqual(len(list(ret.items())), 1)  # exactly one key is found
        key_data = list(ret.items())[0][1]
        try:
            self.assertEqual(key_data["comment"], "github.com")
            self.assertEqual(key_data["enc"], "ssh-rsa")
            self.assertEqual(
                key_data["options"], ['command="/usr/local/lib/ssh-helper"']
            )
            self.assertEqual(key_data["fingerprint"], GITHUB_FINGERPRINT)
        except AssertionError as exc:
            raise AssertionError(f"AssertionError: {exc}. Function returned: {ret}")

    @pytest.mark.slow_test
    def test_bad_enctype(self):
        """
        test to make sure that bad key encoding types don't generate an
        invalid key entry in authorized_keys
        """
        shutil.copyfile(
            os.path.join(RUNTIME_VARS.FILES, "ssh", "authorized_badkeys"),
            self.authorized_keys,
        )
        ret = self.run_function("ssh.auth_keys", ["root", self.authorized_keys])

        # The authorized_badkeys file contains a key with an invalid ssh key
        # encoding (dsa-sha2-nistp256 instead of ecdsa-sha2-nistp256)
        # auth_keys should skip any keys with invalid encodings.  Internally
        # the minion will throw a CommandExecutionError so the
        # user will get an indicator of what went wrong.
        self.assertEqual(len(list(ret.items())), 0)  # Zero keys found

    @pytest.mark.slow_test
    def test_get_known_host_entries(self):
        """
        Check that known host information is returned from ~/.ssh/config
        """
        shutil.copyfile(
            os.path.join(RUNTIME_VARS.FILES, "ssh", "known_hosts"), self.known_hosts
        )
        arg = ["root", "github.com"]
        kwargs = {"config": self.known_hosts}
        ret = self.run_function("ssh.get_known_host_entries", arg, **kwargs)[0]
        try:
            self.assertEqual(ret["enc"], "ssh-rsa")
            self.assertEqual(ret["key"], self.key)
            self.assertEqual(ret["fingerprint"], GITHUB_FINGERPRINT)
        except AssertionError as exc:
            raise AssertionError(f"AssertionError: {exc}. Function returned: {ret}")

    @pytest.mark.skip_on_photonos(
        reason="Skip on PhotonOS.  Attempting to receive the SSH key from Github, using RSA keys which are disabled.",
    )
    @pytest.mark.slow_test
    def test_recv_known_host_entries(self):
        """
        Check that known host information is returned from remote host
        """
        ret = self.run_function(
            "ssh.recv_known_host_entries", ["github.com"], enc="ssh-rsa"
        )
        try:
            self.assertNotEqual(ret, None)
            self.assertEqual(ret[0]["enc"], "ssh-rsa")
            self.assertEqual(ret[0]["key"], self.key)
            self.assertEqual(ret[0]["fingerprint"], GITHUB_FINGERPRINT)
        except AssertionError as exc:
            raise AssertionError(f"AssertionError: {exc}. Function returned: {ret}")

    @pytest.mark.slow_test
    def test_check_known_host_add(self):
        """
        Check known hosts by its fingerprint. File needs to be updated
        """
        arg = ["root", "github.com"]
        kwargs = {"fingerprint": GITHUB_FINGERPRINT, "config": self.known_hosts}
        ret = self.run_function("ssh.check_known_host", arg, **kwargs)
        self.assertEqual(ret, "add")

    @pytest.mark.slow_test
    def test_check_known_host_update(self):
        """
        ssh.check_known_host update verification
        """
        shutil.copyfile(
            os.path.join(RUNTIME_VARS.FILES, "ssh", "known_hosts"), self.known_hosts
        )
        arg = ["root", "github.com"]
        kwargs = {"config": self.known_hosts}
        # wrong fingerprint
        ret = self.run_function(
            "ssh.check_known_host", arg, **dict(kwargs, fingerprint="aa:bb:cc:dd")
        )
        self.assertEqual(ret, "update")
        # wrong keyfile
        ret = self.run_function("ssh.check_known_host", arg, **dict(kwargs, key="YQ=="))
        self.assertEqual(ret, "update")

    @pytest.mark.slow_test
    def test_check_known_host_exists(self):
        """
        Verify check_known_host_exists
        """
        shutil.copyfile(
            os.path.join(RUNTIME_VARS.FILES, "ssh", "known_hosts"), self.known_hosts
        )
        arg = ["root", "github.com"]
        kwargs = {"config": self.known_hosts}
        # wrong fingerprint
        ret = self.run_function(
            "ssh.check_known_host", arg, **dict(kwargs, fingerprint=GITHUB_FINGERPRINT)
        )
        self.assertEqual(ret, "exists")
        # wrong keyfile
        ret = self.run_function(
            "ssh.check_known_host", arg, **dict(kwargs, key=self.key)
        )
        self.assertEqual(ret, "exists")

    @pytest.mark.slow_test
    def test_rm_known_host(self):
        """
        ssh.rm_known_host
        """
        shutil.copyfile(
            os.path.join(RUNTIME_VARS.FILES, "ssh", "known_hosts"), self.known_hosts
        )
        arg = ["root", "github.com"]
        kwargs = {"config": self.known_hosts, "key": self.key}
        # before removal
        ret = self.run_function("ssh.check_known_host", arg, **kwargs)
        self.assertEqual(ret, "exists")
        # remove
        self.run_function("ssh.rm_known_host", arg, config=self.known_hosts)
        # after removal
        ret = self.run_function("ssh.check_known_host", arg, **kwargs)
        self.assertEqual(ret, "add")

    @pytest.mark.skip_on_photonos(
        reason="Skip on PhotonOS.  Attempting to receive the SSH key from Github, using RSA keys which are disabled.",
    )
    @pytest.mark.slow_test
    def test_set_known_host(self):
        """
        ssh.set_known_host
        """
        # add item
        ret = self.run_function(
            "ssh.set_known_host",
            ["root", "github.com"],
            enc="ssh-rsa",
            config=self.known_hosts,
        )
        try:
            self.assertEqual(ret["status"], "updated")
            self.assertEqual(ret["old"], None)
            self.assertEqual(ret["new"][0]["fingerprint"], GITHUB_FINGERPRINT)
        except AssertionError as exc:
            raise AssertionError(f"AssertionError: {exc}. Function returned: {ret}")
        # check that item does exist
        ret = self.run_function(
            "ssh.get_known_host_entries",
            ["root", "github.com"],
            config=self.known_hosts,
        )[0]
        try:
            self.assertEqual(ret["fingerprint"], GITHUB_FINGERPRINT)
        except AssertionError as exc:
            raise AssertionError(f"AssertionError: {exc}. Function returned: {ret}")
        # add the same item once again
        ret = self.run_function(
            "ssh.set_known_host", ["root", "github.com"], config=self.known_hosts
        )
        try:
            self.assertEqual(ret["status"], "exists")
        except AssertionError as exc:
            raise AssertionError(f"AssertionError: {exc}. Function returned: {ret}")



"""
    :codeauthor: :email:`Jayesh Kariya <jayeshk@saltstack.com>`
"""

import pytest

import salt.states.nexus as nexus
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {nexus: {}}


def test_downloaded():
    """
    Test to ensures that the artifact from nexus exists at
    given location.
    """
    name = "jboss"
    arti_url = "http://nexus.example.com/repository"
    artifact = {
        "nexus_url": arti_url,
        "artifact_id": "module",
        "repository": "libs-releases",
        "packaging": "jar",
        "group_id": "com.company.module",
        "classifier": "sources",
        "version": "1.0",
    }

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mck = MagicMock(return_value={"status": False, "changes": {}, "comment": ""})
    with patch.dict(nexus.__salt__, {"nexus.get_release": mck}):
        assert nexus.downloaded(name, artifact) == ret

    with patch.object(
        nexus, "__fetch_from_nexus", MagicMock(side_effect=Exception("error"))
    ):
        ret = nexus.downloaded(name, artifact)
        assert ret["result"] is False
        assert ret["comment"] == "error"



import logging
import os.path
import sys
from copy import deepcopy

import pytest

import salt.config
import salt.loader
import salt.modules.boto_elb as boto_elb
from salt.utils.versions import Version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

# pylint: disable=import-error
try:
    import boto

    boto.ENDPOINTS_PATH = os.path.join(
        RUNTIME_VARS.TESTS_DIR, "unit/files/endpoints.json"
    )
    import boto.ec2.elb

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

try:
    import pkg_resources
    from moto import mock_ec2_deprecated  # pylint: disable=no-name-in-module
    from moto import mock_elb_deprecated  # pylint: disable=no-name-in-module

    HAS_MOTO = True
except ImportError:
    HAS_MOTO = False

    def mock_ec2_deprecated(self):
        """
        if the mock_ec2_deprecated function is not available due to import failure
        this replaces the decorated function with stub_function.
        Allows boto_elb unit tests to use the @mock_ec2_deprecated decorator
        without a "NameError: name 'mock_ec2_deprecated' is not defined" error.
        """

        def stub_function(self):
            pass

        return stub_function

    def mock_elb_deprecated(self):
        """
        if the mock_elb_deprecated function is not available due to import failure
        this replaces the decorated function with stub_function.
        Allows boto_elb unit tests to use the @mock_elb_deprecated decorator
        without a "NameError: name 'mock_elb_deprecated' is not defined" error.
        """

        def stub_function(self):
            pass

        return stub_function


# pylint: enable=import-error


log = logging.getLogger(__name__)

region = "us-east-1"
access_key = "GKTADJGHEIQSXMKKRBJ08H"
secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
conn_parameters = {
    "region": region,
    "key": access_key,
    "keyid": secret_key,
    "profile": {},
}
boto_conn_parameters = {
    "aws_access_key_id": access_key,
    "aws_secret_access_key": secret_key,
}
instance_parameters = {"instance_type": "t1.micro"}
required_moto = "0.3.7"
required_moto_py3 = "1.0.1"


def _has_required_moto():
    """
    Returns True or False depending on if ``moto`` is installed and at the correct version,
    depending on what version of Python is running these tests.
    """
    if not HAS_MOTO:
        return False
    else:
        moto_version = Version(pkg_resources.get_distribution("moto").version)
        if moto_version < Version(required_moto):
            return False
        elif moto_version < Version(required_moto_py3):
            return False

    return True


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_moto() is False,
    reason="The moto module must be >= to {} for PY2 or {} for PY3.".format(
        required_moto, required_moto_py3
    ),
)
class BotoElbTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.boto_elb module
    """

    def setup_loader_modules(self):
        opts = salt.config.DEFAULT_MASTER_OPTS.copy()
        utils = salt.loader.utils(
            opts, whitelist=["boto", "args", "systemd", "path", "platform"]
        )
        funcs = salt.loader.minion_mods(opts, utils=utils)
        return {boto_elb: {"__opts__": opts, "__utils__": utils, "__salt__": funcs}}

    def setUp(self):
        TestCase.setUp(self)
        # __virtual__ must be caller in order for _get_conn to be injected
        boto_elb.__virtual__()

    @mock_ec2_deprecated
    @mock_elb_deprecated
    def test_register_instances_valid_id_result_true(self):
        """
        tests that given a valid instance id and valid ELB that
        register_instances returns True.
        """
        conn_ec2 = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        conn_elb = boto.ec2.elb.connect_to_region(region, **boto_conn_parameters)
        zones = [zone.name for zone in conn_ec2.get_all_zones()]
        elb_name = "TestRegisterInstancesValidIdResult"
        conn_elb.create_load_balancer(elb_name, zones, [(80, 80, "http")])
        reservations = conn_ec2.run_instances("ami-08389d60")
        register_result = boto_elb.register_instances(
            elb_name, reservations.instances[0].id, **conn_parameters
        )
        self.assertEqual(True, register_result)

    @mock_ec2_deprecated
    @mock_elb_deprecated
    def test_register_instances_valid_id_string(self):
        """
        tests that given a string containing a instance id and valid ELB that
        register_instances adds the given instance to an ELB
        """
        conn_ec2 = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        conn_elb = boto.ec2.elb.connect_to_region(region, **boto_conn_parameters)
        zones = [zone.name for zone in conn_ec2.get_all_zones()]
        elb_name = "TestRegisterInstancesValidIdResult"
        conn_elb.create_load_balancer(elb_name, zones, [(80, 80, "http")])
        reservations = conn_ec2.run_instances("ami-08389d60")
        boto_elb.register_instances(
            elb_name, reservations.instances[0].id, **conn_parameters
        )
        load_balancer_refreshed = conn_elb.get_all_load_balancers(elb_name)[0]
        registered_instance_ids = [
            instance.id for instance in load_balancer_refreshed.instances
        ]

        log.debug(load_balancer_refreshed.instances)
        self.assertEqual([reservations.instances[0].id], registered_instance_ids)

    @mock_ec2_deprecated
    @mock_elb_deprecated
    def test_deregister_instances_valid_id_result_true(self):
        """
        tests that given an valid id the boto_elb deregister_instances method
        removes exactly one of a number of ELB registered instances
        """
        conn_ec2 = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        conn_elb = boto.ec2.elb.connect_to_region(region, **boto_conn_parameters)
        zones = [zone.name for zone in conn_ec2.get_all_zones()]
        elb_name = "TestDeregisterInstancesValidIdResult"
        load_balancer = conn_elb.create_load_balancer(
            elb_name, zones, [(80, 80, "http")]
        )
        reservations = conn_ec2.run_instances("ami-08389d60")
        load_balancer.register_instances(reservations.instances[0].id)
        deregister_result = boto_elb.deregister_instances(
            elb_name, reservations.instances[0].id, **conn_parameters
        )
        self.assertEqual(True, deregister_result)

    @mock_ec2_deprecated
    @mock_elb_deprecated
    def test_deregister_instances_valid_id_string(self):
        """
        tests that given an valid id the boto_elb deregister_instances method
        removes exactly one of a number of ELB registered instances
        """
        conn_ec2 = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        conn_elb = boto.ec2.elb.connect_to_region(region, **boto_conn_parameters)
        zones = [zone.name for zone in conn_ec2.get_all_zones()]
        elb_name = "TestDeregisterInstancesValidIdString"
        load_balancer = conn_elb.create_load_balancer(
            elb_name, zones, [(80, 80, "http")]
        )
        reservations = conn_ec2.run_instances("ami-08389d60", min_count=2)
        all_instance_ids = [instance.id for instance in reservations.instances]
        load_balancer.register_instances(all_instance_ids)
        boto_elb.deregister_instances(
            elb_name, reservations.instances[0].id, **conn_parameters
        )
        load_balancer_refreshed = conn_elb.get_all_load_balancers(elb_name)[0]
        expected_instances = deepcopy(all_instance_ids)
        expected_instances.remove(reservations.instances[0].id)
        actual_instances = [
            instance.id for instance in load_balancer_refreshed.instances
        ]
        self.assertEqual(actual_instances, expected_instances)

    @mock_ec2_deprecated
    @mock_elb_deprecated
    def test_deregister_instances_valid_id_list(self):
        """
        tests that given an valid ids in the form of a list that the boto_elb
        deregister_instances all members of the given list
        """
        conn_ec2 = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        conn_elb = boto.ec2.elb.connect_to_region(region, **boto_conn_parameters)
        zones = [zone.name for zone in conn_ec2.get_all_zones()]
        elb_name = "TestDeregisterInstancesValidIdList"
        load_balancer = conn_elb.create_load_balancer(
            elb_name, zones, [(80, 80, "http")]
        )
        reservations = conn_ec2.run_instances("ami-08389d60", min_count=3)
        all_instance_ids = [instance.id for instance in reservations.instances]
        load_balancer.register_instances(all_instance_ids)
        # reservations.instances[:-1] refers to all instances except list
        # instance
        deregister_instances = [instance.id for instance in reservations.instances[:-1]]
        expected_instances = [reservations.instances[-1].id]
        boto_elb.deregister_instances(elb_name, deregister_instances, **conn_parameters)
        load_balancer_refreshed = conn_elb.get_all_load_balancers(elb_name)[0]
        actual_instances = [
            instance.id for instance in load_balancer_refreshed.instances
        ]
        self.assertEqual(actual_instances, expected_instances)

    @mock_ec2_deprecated
    @mock_elb_deprecated
    @pytest.mark.skipif(
        sys.version_info > (3, 6),
        reason="Disabled for 3.7+ pending https://github.com/spulec/moto/issues/1706.",
    )
    def test_get_elb_config(self):
        """
        tests that given an valid ids in the form of a list that the boto_elb
        deregister_instances all members of the given list
        """
        conn_ec2 = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        conn_elb = boto.ec2.elb.connect_to_region(region, **boto_conn_parameters)
        zones = [zone.name for zone in conn_ec2.get_all_zones()]
        elb_name = "TestGetELBConfig"
        load_balancer = conn_elb.create_load_balancer(
            elb_name, zones, [(80, 80, "http")]
        )
        reservations = conn_ec2.run_instances("ami-08389d60", min_count=3)
        all_instance_ids = [instance.id for instance in reservations.instances]
        load_balancer.register_instances(all_instance_ids)

        # DescribeTags does not appear to be included in moto
        # so mock the _get_all_tags function.  Ideally we wouldn't
        # need to mock this.
        with patch("salt.modules.boto_elb._get_all_tags", MagicMock(return_value=None)):
            ret = boto_elb.get_elb_config(elb_name, **conn_parameters)
            _expected_keys = [
                "subnets",
                "availability_zones",
                "canonical_hosted_zone_name_id",
                "tags",
                "dns_name",
                "listeners",
                "backends",
                "policies",
                "vpc_id",
                "scheme",
                "canonical_hosted_zone_name",
                "security_groups",
            ]
            for key in _expected_keys:
                self.assertIn(key, ret)



import pathlib

import pytest

import salt.modules.runit as runit
from tests.support.mock import patch

pytestmark = [pytest.mark.skip_on_windows]


@pytest.fixture
def configure_loader_modules():
    return {runit: {}}


@pytest.fixture
def service_dir(tmp_path):
    dirname = tmp_path / "services"
    dirname.mkdir(exist_ok=True, parents=True)
    return str(dirname)


def test__get_svc_path_on_non_symlinked_service(service_dir):
    service = pathlib.Path(service_dir, "service")
    service.mkdir(exist_ok=True, parents=True)
    service_runfile = service / "run"
    service_runfile.touch()

    with patch.object(runit, "SERVICE_DIR", service_dir):
        with patch("os.access", return_value=True):
            path_list = runit._get_svc_path(str(service), "ENABLED")
            assert path_list
            assert len(path_list) == 1
            assert path_list[0] == str(service)


def test__get_svc_path_on_symlinked_service(service_dir, tmp_path):
    sym_dir = tmp_path / "sym_dir"
    sym_dir.mkdir(exist_ok=True, parents=True)
    service_runfile = sym_dir / "run"
    service_runfile.touch()

    # Create the symlink
    service = pathlib.Path(service_dir, "service")
    service.symlink_to(sym_dir)

    with patch.object(runit, "SERVICE_DIR", service_dir):
        with patch("os.access", return_value=True):
            path_list = runit._get_svc_path(str(service), "ENABLED")
            assert path_list
            assert len(path_list) == 1
            assert path_list[0] == str(sym_dir)



"""
Jinja loading utils to enable a more powerful backend for jinja templates

.. include:: ../../../_incl/jinja_security.rst
"""

import logging
from io import StringIO

import salt.utils.templates
from salt.exceptions import SaltRenderError
from salt.loader.context import NamedLoaderContext

log = logging.getLogger(__name__)


def _split_module_dicts():
    """
    Create a copy of __salt__ dictionary with module.function and module[function]

    Takes advantage of Jinja's syntactic sugar lookup:

    .. code-block::

        {{ salt.cmd.run('uptime') }}
    """
    funcs = __salt__
    if isinstance(__salt__, NamedLoaderContext) and isinstance(__salt__.value(), dict):
        funcs = __salt__.value()
    if not isinstance(funcs, dict):
        return funcs
    mod_dict = dict(funcs)
    for module_func_name, mod_fun in mod_dict.copy().items():
        mod, fun = module_func_name.split(".", 1)
        if mod not in mod_dict:
            # create an empty object that we can add attributes to
            mod_dict[mod] = lambda: None
        setattr(mod_dict[mod], fun, mod_fun)
    return mod_dict


def render(
    template_file,
    saltenv="base",
    sls="",
    argline="",
    context=None,
    tmplpath=None,
    **kws,
):
    """
    Render the template_file, passing the functions and grains into the
    Jinja rendering system.

    :rtype: string
    """
    from_str = argline == "-s"
    if not from_str and argline:
        raise SaltRenderError(f"Unknown renderer option: {argline}")

    tmp_data = salt.utils.templates.JINJA(
        template_file,
        to_str=True,
        salt=_split_module_dicts(),
        grains=__grains__,
        opts=__opts__,
        pillar=__pillar__,
        saltenv=saltenv,
        sls=sls,
        context=context,
        tmplpath=tmplpath,
        proxy=__proxy__,
        from_str=from_str,
        **kws,
    )
    if not tmp_data.get("result", False):
        raise SaltRenderError(
            tmp_data.get("data", "Unknown render error in jinja renderer")
        )
    if isinstance(tmp_data["data"], bytes):
        tmp_data["data"] = tmp_data["data"].decode(__salt_system_encoding__)
    return StringIO(tmp_data["data"])



"""
:codeauthor: Thayne Harbaugh (tharbaug@adobe.com)
"""

import logging

import pytest
from pytestshellutils.exceptions import FactoryNotStarted
from saltfactories.utils import random_string

import salt.defaults.exitcodes
from tests.conftest import FIPS_TESTRUN
from tests.support.helpers import PRE_PYTEST_SKIP_REASON

log = logging.getLogger(__name__)


@pytest.fixture
def proxy_minion_id(salt_master):
    _proxy_minion_id = random_string("proxy-minion-")

    try:
        yield _proxy_minion_id
    finally:
        # Remove stale key if it exists
        pytest.helpers.remove_stale_minion_key(salt_master, _proxy_minion_id)


@pytest.mark.core_test
def test_exit_status_no_proxyid(salt_master, proxy_minion_id):
    """
    Ensure correct exit status when --proxyid argument is missing.
    """
    with pytest.raises(FactoryNotStarted) as exc:
        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id,
            include_proxyid_cli_flag=False,
            overrides={
                "fips_mode": FIPS_TESTRUN,
                "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
                "signing_algorithm": (
                    "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
                ),
            },
        )
        factory.start(start_timeout=10, max_start_attempts=1)

    assert exc.value.process_result.returncode == salt.defaults.exitcodes.EX_USAGE
    assert "Usage" in exc.value.process_result.stderr
    assert "error: salt-proxy requires --proxyid" in exc.value.process_result.stderr


@pytest.mark.skip_on_windows(reason="Windows does not do user checks")
@pytest.mark.core_test
def test_exit_status_unknown_user(salt_master, proxy_minion_id):
    """
    Ensure correct exit status when the proxy is configured to run as an
    unknown user.
    """
    with pytest.raises(FactoryNotStarted) as exc:
        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id,
            overrides={
                "user": "unknown-user",
                "fips_mode": FIPS_TESTRUN,
                "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
                "signing_algorithm": (
                    "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
                ),
            },
        )
        factory.start(start_timeout=10, max_start_attempts=1)

    assert exc.value.process_result.returncode == salt.defaults.exitcodes.EX_NOUSER
    assert "The user is not available." in exc.value.process_result.stderr


@pytest.mark.core_test
def test_exit_status_unknown_argument(salt_master, proxy_minion_id):
    """
    Ensure correct exit status when an unknown argument is passed to
    salt-proxy.
    """
    with pytest.raises(FactoryNotStarted) as exc:
        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id,
            overrides={
                "fips_mode": FIPS_TESTRUN,
                "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
                "signing_algorithm": (
                    "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
                ),
            },
        )
        factory.start("--unknown-argument", start_timeout=10, max_start_attempts=1)

    assert exc.value.process_result.returncode == salt.defaults.exitcodes.EX_USAGE
    assert "Usage" in exc.value.process_result.stderr
    assert "no such option: --unknown-argument" in exc.value.process_result.stderr


# Hangs on Windows. You can add a timeout to the proxy.run command, but then
# it just times out.
@pytest.mark.skip_on_windows(reason=PRE_PYTEST_SKIP_REASON)
def test_exit_status_correct_usage(salt_master, proxy_minion_id, salt_cli):
    """
    Ensure correct exit status when salt-proxy starts correctly.

    Skip on Windows because daemonization not supported
    """
    factory = salt_master.salt_proxy_minion_daemon(
        proxy_minion_id,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
        defaults={"transport": salt_master.config["transport"]},
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    factory.start()
    assert factory.is_running()
    # Let's issue a ping before terminating
    ret = salt_cli.run("test.ping", minion_tgt=proxy_minion_id)
    assert ret.returncode == 0
    assert ret.data is True
    # Terminate the proxy minion
    ret = factory.terminate()
    assert ret.returncode == salt.defaults.exitcodes.EX_OK, ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.redismod
"""

from datetime import datetime

import pytest

import salt.modules.redismod as redismod
from tests.support.mock import MagicMock


class Mockredis:
    """
    Mock redis class
    """

    class ConnectionError(Exception):
        """
        Mock ConnectionError class
        """


class MockConnect:
    """
    Mock Connect class
    """

    counter = 0

    def __init__(self):
        self.name = None
        self.pattern = None
        self.value = None
        self.key = None
        self.seconds = None
        self.timestamp = None
        self.field = None
        self.start = None
        self.stop = None
        self.master_host = None
        self.master_port = None

    @staticmethod
    def bgrewriteaof():
        """
        Mock bgrewriteaof method
        """
        return "A"

    @staticmethod
    def bgsave():
        """
        Mock bgsave method
        """
        return "A"

    def config_get(self, pattern):
        """
        Mock config_get method
        """
        self.pattern = pattern
        return "A"

    def config_set(self, name, value):
        """
        Mock config_set method
        """
        self.name = name
        self.value = value
        return "A"

    @staticmethod
    def dbsize():
        """
        Mock dbsize method
        """
        return "A"

    @staticmethod
    def delete():
        """
        Mock delete method
        """
        return "A"

    def exists(self, key):
        """
        Mock exists method
        """
        self.key = key
        return "A"

    def expire(self, key, seconds):
        """
        Mock expire method
        """
        self.key = key
        self.seconds = seconds
        return "A"

    def expireat(self, key, timestamp):
        """
        Mock expireat method
        """
        self.key = key
        self.timestamp = timestamp
        return "A"

    @staticmethod
    def flushall():
        """
        Mock flushall method
        """
        return "A"

    @staticmethod
    def flushdb():
        """
        Mock flushdb method
        """
        return "A"

    def get(self, key):
        """
        Mock get method
        """
        self.key = key
        return "A"

    def hget(self, key, field):
        """
        Mock hget method
        """
        self.key = key
        self.field = field
        return "A"

    def hgetall(self, key):
        """
        Mock hgetall method
        """
        self.key = key
        return "A"

    @staticmethod
    def info():
        """
        Mock info method
        """
        return "A"

    def keys(self, pattern):
        """
        Mock keys method
        """
        self.pattern = pattern
        return "A"

    def type(self, key):
        """
        Mock type method
        """
        self.key = key
        return "A"

    @staticmethod
    def lastsave():
        """
        Mock lastsave method
        """
        return datetime.now()

    def llen(self, key):
        """
        Mock llen method
        """
        self.key = key
        return "A"

    def lrange(self, key, start, stop):
        """
        Mock lrange method
        """
        self.key = key
        self.start = start
        self.stop = stop
        return "A"

    @staticmethod
    def ping():
        """
        Mock ping method
        """
        MockConnect.counter = MockConnect.counter + 1
        if MockConnect.counter == 1:
            return "A"
        elif MockConnect.counter in (2, 3, 5):
            raise Mockredis.ConnectionError("foo")

    @staticmethod
    def save():
        """
        Mock save method
        """
        return "A"

    def set(self, key, value):
        """
        Mock set method
        """
        self.key = key
        self.value = value
        return "A"

    @staticmethod
    def shutdown():
        """
        Mock shutdown method
        """
        return "A"

    def slaveof(self, master_host, master_port):
        """
        Mock slaveof method
        """
        self.master_host = master_host
        self.master_port = master_port
        return "A"

    def smembers(self, key):
        """
        Mock smembers method
        """
        self.key = key
        return "A"

    @staticmethod
    def time():
        """
        Mock time method
        """
        return "A"

    def zcard(self, key):
        """
        Mock zcard method
        """
        self.key = key
        return "A"

    def zrange(self, key, start, stop):
        """
        Mock zrange method
        """
        self.key = key
        self.start = start
        self.stop = stop
        return "A"


@pytest.fixture
def configure_loader_modules():
    return {
        redismod: {
            "redis": Mockredis,
            "_connect": MagicMock(return_value=MockConnect()),
        }
    }


def test_bgrewriteaof():
    """
    Test to asynchronously rewrite the append-only file
    """
    assert redismod.bgrewriteaof() == "A"


def test_bgsave():
    """
    Test to asynchronously save the dataset to disk
    """
    assert redismod.bgsave() == "A"


def test_config_get():
    """
    Test to get redis server configuration values
    """
    assert redismod.config_get("*") == "A"


def test_config_set():
    """
    Test to set redis server configuration values
    """
    assert redismod.config_set("name", "value") == "A"


def test_dbsize():
    """
    Test to return the number of keys in the selected database
    """
    assert redismod.dbsize() == "A"


def test_delete():
    """
    Test to deletes the keys from redis, returns number of keys deleted
    """
    assert redismod.delete() == "A"


def test_exists():
    """
    Test to return true if the key exists in redis
    """
    assert redismod.exists("key") == "A"


def test_expire():
    """
    Test to set a keys time to live in seconds
    """
    assert redismod.expire("key", "seconds") == "A"


def test_expireat():
    """
    Test to set a keys expire at given UNIX time
    """
    assert redismod.expireat("key", "timestamp") == "A"


def test_flushall():
    """
    Test to remove all keys from all databases
    """
    assert redismod.flushall() == "A"


def test_flushdb():
    """
    Test to remove all keys from the selected database
    """
    assert redismod.flushdb() == "A"


def test_get_key():
    """
    Test to get redis key value
    """
    assert redismod.get_key("key") == "A"


def test_hget():
    """
    Test to get specific field value from a redis hash, returns dict
    """
    assert redismod.hget("key", "field") == "A"


def test_hgetall():
    """
    Test to get all fields and values from a redis hash, returns dict
    """
    assert redismod.hgetall("key") == "A"


def test_info():
    """
    Test to get information and statistics about the server
    """
    assert redismod.info() == "A"


def test_keys():
    """
    Test to get redis keys, supports glob style patterns
    """
    assert redismod.keys("pattern") == "A"


def test_key_type():
    """
    Test to get redis key type
    """
    assert redismod.key_type("key") == "A"


def test_lastsave():
    """
    Test to get the UNIX time in seconds of the last successful
    save to disk
    """
    assert redismod.lastsave()


def test_llen():
    """
    Test to get the length of a list in Redis
    """
    assert redismod.llen("key") == "A"


def test_lrange():
    """
    Test to get a range of values from a list in Redis
    """
    assert redismod.lrange("key", "start", "stop") == "A"


def test_ping():
    """
    Test to ping the server, returns False on connection errors
    """
    assert redismod.ping() == "A"

    assert not redismod.ping()


def test_save():
    """
    Test to synchronously save the dataset to disk
    """
    assert redismod.save() == "A"


def test_set_key():
    """
    Test to set redis key value
    """
    assert redismod.set_key("key", "value") == "A"


def test_shutdown():
    """
    Test to synchronously save the dataset to disk and then
    shut down the server
    """
    assert not redismod.shutdown()

    assert redismod.shutdown()

    assert not redismod.shutdown()


def test_slaveof():
    """
    Test to make the server a slave of another instance, or
    promote it as master
    """
    assert redismod.slaveof("master_host", "master_port") == "A"


def test_smembers():
    """
    Test to get members in a Redis set
    """
    assert redismod.smembers("key") == ["A"]


def test_time():
    """
    Test to return the current server UNIX time in seconds
    """
    assert redismod.time() == "A"


def test_zcard():
    """
    Test to get the length of a sorted set in Redis
    """
    assert redismod.zcard("key") == "A"


def test_zrange():
    """
    Test to get a range of values from a sorted set in Redis by index
    """
    assert redismod.zrange("key", "start", "stop") == "A"



"""
Tests for the file state
"""

import errno
import os
import tempfile
import textwrap
import time

import pytest

import salt.utils.files
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.runtests import RUNTIME_VARS


class CMDTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the cmd state
    """

    @classmethod
    def setUpClass(cls):
        cls.__cmd = "dir" if salt.utils.platform.is_windows() else "ls"

    def test_run_simple(self):
        """
        cmd.run
        """
        ret = self.run_state("cmd.run", name=self.__cmd, cwd=tempfile.gettempdir())
        self.assertSaltTrueReturn(ret)

    def test_run_output_loglevel(self):
        """
        cmd.run with output_loglevel=quiet
        """
        ret = self.run_state(
            "cmd.run",
            name=self.__cmd,
            cwd=tempfile.gettempdir(),
            output_loglevel="quiet",
        )
        self.assertSaltTrueReturn(ret)

    def test_run_simple_test_true(self):
        """
        cmd.run test interface
        """
        ret = self.run_state(
            "cmd.run", name=self.__cmd, cwd=tempfile.gettempdir(), test=True
        )
        self.assertSaltNoneReturn(ret)

    def test_run_hide_output(self):
        """
        cmd.run with output hidden
        """
        ret = self.run_state("cmd.run", name=self.__cmd, hide_output=True)
        self.assertSaltTrueReturn(ret)
        ret = ret[next(iter(ret))]
        self.assertEqual(ret["changes"]["stdout"], "")
        self.assertEqual(ret["changes"]["stderr"], "")


class CMDRunRedirectTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the cmd state of run_redirect
    """

    def setUp(self):
        self.state_name = "run_redirect"
        state_filename = self.state_name + ".sls"
        self.state_file = os.path.join(RUNTIME_VARS.TMP_STATE_TREE, state_filename)

        # Create the testfile and release the handle
        fd, self.test_file = tempfile.mkstemp()
        try:
            os.close(fd)
        except OSError as exc:
            if exc.errno != errno.EBADF:
                raise

        # Create the testfile and release the handle
        fd, self.test_tmp_path = tempfile.mkstemp()
        try:
            os.close(fd)
        except OSError as exc:
            if exc.errno != errno.EBADF:
                raise

        super().setUp()

    def tearDown(self):
        time.sleep(1)
        for path in (self.state_file, self.test_tmp_path, self.test_file):
            try:
                os.remove(path)
            except OSError:
                # Not all of the tests leave files around that we want to remove
                # As some of the tests create the sls files in the test itself,
                # And some are using files in the integration test file state tree.
                pass
        super().tearDown()

    @pytest.mark.slow_test
    def test_run_unless(self):
        """
        test cmd.run unless
        """
        state_key = "cmd_|-{0}_|-{0}_|-run".format(self.test_tmp_path)
        with salt.utils.files.fopen(self.state_file, "w") as fb_:
            fb_.write(
                salt.utils.stringutils.to_str(
                    textwrap.dedent(
                        """
                {}:
                  cmd.run:
                    - unless: echo cheese > {}
                """.format(
                            self.test_tmp_path, self.test_file
                        )
                    )
                )
            )

        ret = self.run_function("state.sls", [self.state_name])
        self.assertTrue(ret[state_key]["result"])

    @pytest.mark.slow_test
    @pytest.mark.windows_whitelisted
    def test_run_unless_multiple_cmds(self):
        """
        test cmd.run using multiple unless options where the first cmd in the
        list will pass, but the second will fail. This tests the fix for issue
        #35384. (The fix is in PR #35545.)
        """
        sls = self.run_function("state.sls", mods="issue-35384")
        self.assertSaltTrueReturn(sls)
        # We must assert against the comment here to make sure the comment reads that the
        # command "echo "hello"" was run. This ensures that we made it to the last unless
        # command in the state. If the comment reads "unless condition is true", or similar,
        # then the unless state run bailed out after the first unless command succeeded,
        # which is the bug we're regression testing for.
        self.assertEqual(
            sls['cmd_|-cmd_run_unless_multiple_|-echo "hello"_|-run']["comment"],
            'Command "echo "hello"" run',
        )

    @pytest.mark.slow_test
    def test_run_creates_exists(self):
        """
        test cmd.run creates already there
        """
        state_key = "cmd_|-echo >> {0}_|-echo >> {0}_|-run".format(self.test_file)
        with salt.utils.files.fopen(self.state_file, "w") as fb_:
            fb_.write(
                salt.utils.stringutils.to_str(
                    textwrap.dedent(
                        """
                echo >> {0}:
                  cmd.run:
                    - creates: {0}
                """.format(
                            self.test_file
                        )
                    )
                )
            )

        ret = self.run_function("state.sls", [self.state_name])
        self.assertTrue(ret[state_key]["result"])
        self.assertEqual(len(ret[state_key]["changes"]), 0)

    @pytest.mark.slow_test
    def test_run_creates_new(self):
        """
        test cmd.run creates not there
        """
        os.remove(self.test_file)
        state_key = "cmd_|-echo >> {0}_|-echo >> {0}_|-run".format(self.test_file)
        with salt.utils.files.fopen(self.state_file, "w") as fb_:
            fb_.write(
                salt.utils.stringutils.to_str(
                    textwrap.dedent(
                        """
                echo >> {0}:
                  cmd.run:
                    - creates: {0}
                """.format(
                            self.test_file
                        )
                    )
                )
            )

        ret = self.run_function("state.sls", [self.state_name])
        self.assertTrue(ret[state_key]["result"])
        self.assertEqual(len(ret[state_key]["changes"]), 4)

    @pytest.mark.slow_test
    def test_run_redirect(self):
        """
        test cmd.run with shell redirect
        """
        state_key = "cmd_|-echo test > {0}_|-echo test > {0}_|-run".format(
            self.test_file
        )
        with salt.utils.files.fopen(self.state_file, "w") as fb_:
            fb_.write(
                salt.utils.stringutils.to_str(
                    textwrap.dedent(
                        """
                echo test > {}:
                  cmd.run
                """.format(
                            self.test_file
                        )
                    )
                )
            )

        ret = self.run_function("state.sls", [self.state_name])
        self.assertTrue(ret[state_key]["result"])


class CMDRunWatchTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the cmd state of run_watch
    """

    def setUp(self):
        self.state_name = "run_watch"
        state_filename = self.state_name + ".sls"
        self.state_file = os.path.join(RUNTIME_VARS.TMP_STATE_TREE, state_filename)
        super().setUp()

    def tearDown(self):
        os.remove(self.state_file)
        super().tearDown()

    def test_run_watch(self):
        """
        test cmd.run watch
        """
        saltines_key = "cmd_|-saltines_|-echo changed=true_|-run"
        biscuits_key = "cmd_|-biscuits_|-echo biscuits_|-wait"

        with salt.utils.files.fopen(self.state_file, "w") as fb_:
            fb_.write(
                salt.utils.stringutils.to_str(
                    textwrap.dedent(
                        """
                saltines:
                  cmd.run:
                    - name: echo changed=true
                    - cwd: /
                    - stateful: True

                biscuits:
                  cmd.wait:
                    - name: echo biscuits
                    - cwd: /
                    - watch:
                        - cmd: saltines
                """
                    )
                )
            )

        ret = self.run_function("state.sls", [self.state_name])
        self.assertTrue(ret[saltines_key]["result"])
        self.assertTrue(ret[biscuits_key]["result"])



import shutil
import subprocess

import psutil
import pytest

try:
    import gnupg as gnupglib

    HAS_GNUPG = True
except ImportError:
    HAS_GNUPG = False


pytestmark = [
    pytest.mark.skipif(HAS_GNUPG is False, reason="Needs python-gnupg library"),
    pytest.mark.skip_if_binaries_missing("gpg", reason="Needs gpg binary"),
]


@pytest.fixture
def gpghome(tmp_path):
    root = tmp_path / "gpghome"
    root.mkdir(mode=0o0700)
    try:
        yield root
    finally:
        # Make sure we don't leave any gpg-agents running behind
        gpg_connect_agent = shutil.which("gpg-connect-agent")
        if gpg_connect_agent:
            gnupghome = root / ".gnupg"
            if not gnupghome.is_dir():
                gnupghome = root
            try:
                subprocess.run(
                    [gpg_connect_agent, "killagent", "/bye"],
                    env={"GNUPGHOME": str(gnupghome)},
                    shell=False,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            except subprocess.CalledProcessError:
                # This is likely CentOS 7 or Amazon Linux 2
                pass

        # If the above errored or was not enough, as a last resort, let's check
        # the running processes.
        for proc in psutil.process_iter():
            try:
                if "gpg-agent" in proc.name():
                    for arg in proc.cmdline():
                        if str(root) in arg:
                            proc.terminate()
            except Exception:  # pylint: disable=broad-except
                pass


@pytest.fixture
def gpg(loaders, modules, gpghome):
    yield modules.gpg


@pytest.fixture
def signed_data(tmp_path):
    signed_data = "What do you have if you have NaCl and NiCd? A salt and battery.\n"
    data = tmp_path / "signed_data"
    data.write_text(signed_data)
    assert data.read_bytes() == signed_data.encode()
    yield data
    data.unlink()


@pytest.fixture
def key_a_fp():
    return "EF03765F59EE904930C8A781553A82A058C0C795"


@pytest.fixture
def key_a_pub():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY4fxHQEEAJvXEaaw+o/yZCwMOJbt5FQHbVMMDX/0YI8UdzsE5YCC4iKnoC3x
FwFdkevKj3qp+45iBGLLnalfXIcVGXJGACB+tPHgsfHaXSDQPSfmX6jbZ6pHosSm
v1tTixY+NTJzGL7hDLz2sAXTbYmTbXeE9ifWWk6NcIwZivUbhNRBM+KxABEBAAG0
LUtleSBBIChHZW5lcmF0ZWQgYnkgU2FsdFN0YWNrKSA8a2V5YUBleGFtcGxlPojR
BBMBCAA7FiEE7wN2X1nukEkwyKeBVTqCoFjAx5UFAmOH8R0CGy8FCwkIBwICIgIG
FQoJCAsCBBYCAwECHgcCF4AACgkQVTqCoFjAx5XURAQAguOwI+49lG0Kby+Bsyv3
of3GgxvhS1Qa7+ysj088az5GVt0pqVe3SbRVvn/jyC6yZvWuv94KdL3R7hCeEz2/
JakCRJ4wxEsdeASE8t9H/oTqD0I5asMa9EMvn5ICEGeLsTeQb7OYYihTQj7HJLG6
pDEmK8EhJDvV/9o0lnhm/9w=
=Wc0O
-----END PGP PUBLIC KEY BLOCK-----"""


@pytest.fixture
def key_a_priv():
    return """\
-----BEGIN PGP PRIVATE KEY BLOCK-----

lQHYBGOH8R0BBACb1xGmsPqP8mQsDDiW7eRUB21TDA1/9GCPFHc7BOWAguIip6At
8RcBXZHryo96qfuOYgRiy52pX1yHFRlyRgAgfrTx4LHx2l0g0D0n5l+o22eqR6LE
pr9bU4sWPjUycxi+4Qy89rAF022Jk213hPYn1lpOjXCMGYr1G4TUQTPisQARAQAB
AAP7BlQ9nKcZI/24hQPxi+qpMGL1VQ87IKBWiBURExHrtrSKFdV4N0lwcV8hGSIK
wfTzmRigvDjwBCQR9E/+brJKWLdGmmHjYHIU3m4fz26E4UlxEu2XfxZOSPKPTnzh
GqVSjmZ9TDdr5Ykpz5SyQ1YOUS9iRI6O5Dp0c4+6n2gyTYECAMQPCa8UnoHw1jgw
JHnK+XM3jinqgIOMS66i5nCGe3PItaAOvPIwA0lyl2Io06lGuiSVIqbJIUTsf2Mv
y14eJnECAMt8O6gMsjJdZ/dU9srqz4ZatPUHtQm2KBnvk311PmeErJ1FiiAqXTVq
Q9y3GvkEnENeuC/ac0XztiHsEC2eIEEB/iu1i5sP3zUZZnBNDbsmDZEy+HKHm8lL
Vg1+hHUdznMMmJ/PKq+WlB3KvdNzhEFd+0R+ylfRTMWnhNMWxL1atNyYC7QtS2V5
IEEgKEdlbmVyYXRlZCBieSBTYWx0U3RhY2spIDxrZXlhQGV4YW1wbGU+iNEEEwEI
ADsWIQTvA3ZfWe6QSTDIp4FVOoKgWMDHlQUCY4fxHQIbLwULCQgHAgIiAgYVCgkI
CwIEFgIDAQIeBwIXgAAKCRBVOoKgWMDHldREBACC47Aj7j2UbQpvL4GzK/eh/caD
G+FLVBrv7KyPTzxrPkZW3SmpV7dJtFW+f+PILrJm9a6/3gp0vdHuEJ4TPb8lqQJE
njDESx14BITy30f+hOoPQjlqwxr0Qy+fkgIQZ4uxN5Bvs5hiKFNCPscksbqkMSYr
wSEkO9X/2jSWeGb/3A==
=lVXx
-----END PGP PRIVATE KEY BLOCK-----"""


@pytest.fixture
def key_a_sig():
    return """\
-----BEGIN PGP SIGNATURE-----

iLMEAAEIAB0WIQTvA3ZfWe6QSTDIp4FVOoKgWMDHlQUCY4fz7QAKCRBVOoKgWMDH
lYH/A/9iRT4JqzZEfRGM6XUyVbLCoF2xyEc43yL06nxn1sMFjMdLOLUDsaZbppHB
vEo4XJ45+Xqu1h3RK6bDVdDURyacPzdef8v357rJHjT8tb68JcwSzCBXLh29Nasl
cOm6/gNVIVm3Uoe9mKXESRvpMYmNUp2UM7ZzzBstK/ViVR82UA==
=EZPV
-----END PGP SIGNATURE-----"""


@pytest.fixture
def key_b_fp():
    return "118B4FAB78038CB2DF7B69E20F6C422647465C93"


@pytest.fixture
def key_b_pub():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY4fxNQEEAOgAzbpheJrOq4il5BrMVtP1G1kU94QX2+xLXEgW/wPdE4HD6Zbg
vliIg18v7Na4x8ubWy/7CkXC83EJ8SoSqcCccvuKjIWsm6tfeCidNstNCjewFMUR
7ZOQmAe/I2JAlz2SgNxS3ZDiCZpGkxqE0GZ+1N7Mz2WHImnExG149RVHABEBAAG0
LUtleSBCIChHZW5lcmF0ZWQgYnkgU2FsdFN0YWNrKSA8a2V5YkBleGFtcGxlPojR
BBMBCAA7FiEEEYtPq3gDjLLfe2niD2xCJkdGXJMFAmOH8TUCGy8FCwkIBwICIgIG
FQoJCAsCBBYCAwECHgcCF4AACgkQD2xCJkdGXJNR3AQAk5ZoN+/ViIX3vA/LbXPn
2VE1E7ETTeIGqsb5f98UfjIbYfkNE8+OtnPxnDbSOPWBEOT+XPPjmxnE0a2UNTfn
ECO71/ZUiyC3ZN50IZ0vgzwBH+DeIV6PDAAun5FGx4RI7v6n0CPlrUcWKYe8wY1F
COflOxnEyLVHXnX8wUIzZwo=
=Hq0X
-----END PGP PUBLIC KEY BLOCK-----"""


@pytest.fixture
def key_b_priv():
    return """\
-----BEGIN PGP PRIVATE KEY BLOCK-----

lQHYBGOH8TUBBADoAM26YXiazquIpeQazFbT9RtZFPeEF9vsS1xIFv8D3ROBw+mW
4L5YiINfL+zWuMfLm1sv+wpFwvNxCfEqEqnAnHL7ioyFrJurX3gonTbLTQo3sBTF
Ee2TkJgHvyNiQJc9koDcUt2Q4gmaRpMahNBmftTezM9lhyJpxMRtePUVRwARAQAB
AAP5ARivbnABmWhchYCiwbnjKSljNvosHJ5seCaUIq34squWaRGH5rBccJiPrQvD
qJvmnWCp+CjNZTysH/eUcz+6xmoefzsB9YPfyWKYCM6yNaBbFceVER+TCRoLjlZi
Nw7VtD6t8Rrxq9T7fykmNU7PD1hOKEXacaHN04jCFtPsXdkCAPK9LkE3c9nVNxUS
17AtviMm2/9nC0wu+9W/Tub0VZjtsanwTyA87uwNlxrOyUeopZZrzSioJEa6b/Hq
RtUmvR8CAPSteh9vXM9D94jpmWzz2Hx1uo8OCDk86KovbEFLqNTEnnbaG8eXOH68
Y0BPKrLRY9lY1zkeudw+pPyg6Q1CetkB/0VbGbfswR15WkiAJnd/Cln07upDUYLD
bQ4C/w7YbfQndmnJ84gHqQH2T4xgULlEvl3+rg+x7vg4z1X7St31hD+kqLQtS2V5
IEIgKEdlbmVyYXRlZCBieSBTYWx0U3RhY2spIDxrZXliQGV4YW1wbGU+iNEEEwEI
ADsWIQQRi0+reAOMst97aeIPbEImR0ZckwUCY4fxNQIbLwULCQgHAgIiAgYVCgkI
CwIEFgIDAQIeBwIXgAAKCRAPbEImR0Zck1HcBACTlmg379WIhfe8D8ttc+fZUTUT
sRNN4gaqxvl/3xR+Mhth+Q0Tz462c/GcNtI49YEQ5P5c8+ObGcTRrZQ1N+cQI7vX
9lSLILdk3nQhnS+DPAEf4N4hXo8MAC6fkUbHhEju/qfQI+WtRxYph7zBjUUI5+U7
GcTItUdedfzBQjNnCg==
=kzV7
-----END PGP PRIVATE KEY BLOCK-----"""


@pytest.fixture
def key_b_sig():
    return """\
-----BEGIN PGP SIGNATURE-----

iLMEAAEIAB0WIQQRi0+reAOMst97aeIPbEImR0ZckwUCY4f0gQAKCRAPbEImR0Zc
kwj5A/9Fj7JFdzu5CGvB7MGYPPUm7cBh231QuSLndlSFvv3ZBrvU9906fstzjQRB
6x4m1uc04bckEoZ/2WE6r5dP39sdvzlyMhGq5qAJZZiuiZF2EjzMnQ/dJoXYM8pM
674JLQ2VjsWCZjp7s3JnPh6+0VL9IoDx3QsfPSBKBjOHOSj0SA==
=1sKz
-----END PGP SIGNATURE-----"""


@pytest.fixture
def key_c_fp():
    return "96F136AC4C92D78DAF33105E35C03186001C6E31"


@pytest.fixture
def key_c_pub():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY4f2GgEEALToT23wZfLGM/JGCV4pWRlIXXqLwwEBSXral92HvsUjC8Vqsh1z
1n0K8/vIpS9OH2Q21emtht4y36rbahy+w6wRc1XXjPQ28Pyd+8v/jSKy/NKW3g+y
ZoB22vj4L35pAu/G6xs9+pKsLHGjMo+LsWZNEZ2Ar06aYA0dbTb0AqYfABEBAAG0
LUtleSBDIChHZW5lcmF0ZWQgYnkgU2FsdFN0YWNrKSA8a2V5Y0BleGFtcGxlPojR
BBMBCAA7FiEElvE2rEyS142vMxBeNcAxhgAcbjEFAmOH9hoCGy8FCwkIBwICIgIG
FQoJCAsCBBYCAwECHgcCF4AACgkQNcAxhgAcbjH2WAP/RtlUfN/novwmxxma6Zom
P6skFnCcRCs0vMU3OnNwuxZt9B+j0sUTu6noGi04Gcbd0eQs7v57DQHcRhNidZU/
8BJv5jD6E2yuzLK9lON+Yhgc6Pg6raA3hBeCY2HuzTEQLAThyV7ihboNILo7FJwo
y9KvnTFP2+oeDX2Z/m4SoWw=
=81Kb
-----END PGP PUBLIC KEY BLOCK-----"""


# the signature was made on different data
@pytest.fixture
def key_c_fail_sig():
    return """\
-----BEGIN PGP SIGNATURE-----

iLMEAAEIAB0WIQSW8TasTJLXja8zEF41wDGGABxuMQUCY4f2pwAKCRA1wDGGABxu
MdtOA/0ROhdh1VgfZn94PWgcrlurwYE0ftI3AXbSI00FsrWviF8WZQtp7YjrUC3t
/a/P1UYZkbBR0EeKwbKGmersq8mfif1qejWpQXpmmx011KQ0AFl7vdZUtcj2n454
KQE7MX8ePBchpSyhmvXzEezyw2FJp+YVOUKhDsc1vbNWTxGYJw==
=6tkK
-----END PGP SIGNATURE-----"""


@pytest.fixture
def key_d_fp():
    return "0B33DC666FA1211311850D4F5F3CAF4AE7BF949D"


@pytest.fixture
def key_d_pub():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY4gB/AEEAK2iwXXY4rbykGMFsLxdcnUdWtglKBXyQl5Y2Mj71UedgGTZVNWP
4dvUnQNqthPCu3hFyD2I+USa80OuYw5k9MPMqZcu3xmJKneaEX07r829G2T8yllz
vgKd2+4aUZHsKFz9fId7lraCDhSpmS0zLeTab6jNDfYecVJB8JaQpJrNABEBAAG0
LUtleSBEIChHZW5lcmF0ZWQgYnkgU2FsdFN0YWNrKSA8a2V5ZEBleGFtcGxlPojR
BBMBCAA7FiEECzPcZm+hIRMRhQ1PXzyvSue/lJ0FAmOIAfwCGy8FCwkIBwICIgIG
FQoJCAsCBBYCAwECHgcCF4AACgkQXzyvSue/lJ373AQAqKGhEfZ8Kx0s7L+T+6Ug
orxgvvQhbcc+eciA9alAptstkt9qg03yaii1InwJE5HjLNqOu8DWnhuBDTquWQUm
CWKUEicI6eZCFHVrjAcUKJcp/Lne9Ny/r1+14eE2vlsMf3oy0n0OpOTC+KhJzDpz
Mq2tZaSibjzjqQ32LiNJi3c=
=rjff
-----END PGP PUBLIC KEY BLOCK-----"""


@pytest.fixture
def key_d_sig():
    return """\
-----BEGIN PGP SIGNATURE-----

iLMEAAEIAB0WIQQLM9xmb6EhExGFDU9fPK9K57+UnQUCY4gCXwAKCRBfPK9K57+U
nYk/BACe4YUPaxHRuky8ya7k27P+RTynbuUGJ1w3AHaO1VXDTvoNTioyaqeDpO6C
NsMl5bIa7GkYEaeM/OPWQ9WlWPno6+nwOEL7/818J+JgxWVsLtGJNlYzuEwI7tDV
FUXLma7dvOw13LLY2RjtgaVbePTO+H+uCrAa4/O0YEIAE2C2fA==
=3m4o
-----END PGP SIGNATURE-----"""


@pytest.fixture
def key_e_fp():
    return "2401C402776328D78D6B4C5D67D35BC98502D9B9"


# expires 2022-12-01
@pytest.fixture
def key_e_pub():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY4gjEQEEAKYpWezZQWiAUDvAMcMhBkjHGY2fM4MMiXc6+fRbNV4VCL9TtJYE
gjccYVu44DtIYQzMVimrPQ6xepUmFRalezCG0OO4v25Ciwyeg8LX+Tb3kyAYFAxi
qLXAJyr3aZ/539xBak/Vf5xdURIi7WF5qBGQxd87tRDDqyPFnr87JJtFABEBAAG0
LUtleSBFIChHZW5lcmF0ZWQgYnkgU2FsdFN0YWNrKSA8a2V5ZUBleGFtcGxlPojX
BBMBCABBFiEEJAHEAndjKNeNa0xdZ9NbyYUC2bkFAmOIIxECGy8FCQAAZh8FCwkI
BwICIgIGFQoJCAsCBBYCAwECHgcCF4AACgkQZ9NbyYUC2bmn1QP/WPVhj1bC9/9R
hifv29MG9maRNIkuEkZKtRJj7HMSaamD5IOtGoyMuBwicb38n2Z2KQZUiJbvyZTt
PS328F8YSUSyWQKqmhwL0iLlnDzx8l/nFr5tiss2b/ZzjlMP4iXtAgEdVMJnfjrM
J7xvL0cNSsHha4hUIrekvzM+SNwYkzs=
=nue4
-----END PGP PUBLIC KEY BLOCK-----"""


@pytest.fixture
def key_e_priv():
    return """\
-----BEGIN PGP PRIVATE KEY BLOCK-----

lQHYBGOIIxEBBACmKVns2UFogFA7wDHDIQZIxxmNnzODDIl3Ovn0WzVeFQi/U7SW
BII3HGFbuOA7SGEMzFYpqz0OsXqVJhUWpXswhtDjuL9uQosMnoPC1/k295MgGBQM
Yqi1wCcq92mf+d/cQWpP1X+cXVESIu1heagRkMXfO7UQw6sjxZ6/OySbRQARAQAB
AAP/Se3mE8qKHpvQlvUpbt83s5PaW7e0rJ8cXo8//SfDs+t5696rX3/8C9c1viCg
q9/FRnN39qw4y1vN5aR/B4dzKmlycWZ5Kk5NewRfNAyvD85uXA2VDDTCZQ7qDlwZ
A24s95xjWWC2urhMszYgWbT9Kzk4Q1aNhPIEs1QgYvIujjUCAMX8+w5FRrOZWYP+
K8qL9ZPUAbUIdg+dCN1rTQvrCbM2FtZquFw89EyzT+qmBkfHCKpczC0XHdqnhsZU
sGTdqXcCANbZEFCNldv6FwhYsDBitP48shoGk3SIsAIpzP3klz+3Eut6LCTzS5T4
t9OVEIpKCP5Ci5m/BoTO/rbKhKt5ECMB/14pYNJcKtolEwvE6w5j6h5f72N+3dTs
qGTpDjfpgrSTmpoZcqDcFg0ycfwFyDzBVI+Hn+6njc/FtMxH3JAA2kug6LQtS2V5
IEUgKEdlbmVyYXRlZCBieSBTYWx0U3RhY2spIDxrZXllQGV4YW1wbGU+iNcEEwEI
AEEWIQQkAcQCd2Mo141rTF1n01vJhQLZuQUCY4gjEQIbLwUJAABmHwULCQgHAgIi
AgYVCgkICwIEFgIDAQIeBwIXgAAKCRBn01vJhQLZuafVA/9Y9WGPVsL3/1GGJ+/b
0wb2ZpE0iS4SRkq1EmPscxJpqYPkg60ajIy4HCJxvfyfZnYpBlSIlu/JlO09Lfbw
XxhJRLJZAqqaHAvSIuWcPPHyX+cWvm2KyzZv9nOOUw/iJe0CAR1Uwmd+OswnvG8v
Rw1KweFriFQit6S/Mz5I3BiTOw==
=49zT
-----END PGP PRIVATE KEY BLOCK-----"""


@pytest.fixture
def key_e_exp_sig():
    return """\
-----BEGIN PGP SIGNATURE-----

iLMEAAEIAB0WIQQkAcQCd2Mo141rTF1n01vJhQLZuQUCY4gjPgAKCRBn01vJhQLZ
ueNkBACa2JjXZjatAbW4wXZ6RKqgluNwvC12EzFbQK2oGh1JC9htS4mVYE2aroup
E5XNzlr8R1z9b5HmuqqSniTNXG5xpxDLvD+rZK3Ww6AIs5FR2R9e0Kr0kMayQFdO
UDJsV8u8GFP+76oGy4G4GU1NlcqWpmIAzst0G4q4ipx891Z10A==
=G8tZ
-----END PGP SIGNATURE-----"""


@pytest.fixture
def key_f_fp():
    return "D7D3F617EBD2113059914876F7D0B975BC8E7ED1"


# no pubkey for this one
@pytest.fixture
def key_f_sig():
    return """\
-----BEGIN PGP SIGNATURE-----

iLMEAAEIAB0WIQTX0/YX69IRMFmRSHb30Ll1vI5+0QUCY4it3wAKCRD30Ll1vI5+
0Qf8A/4yI5nX0Q7kfR6rHvcDgB9HFlimiE15HOd8pW0RWK0YWvTtn+5vWyOZalS/
oHvSiXZCYIgjllvMSJbgxj22GrH621B5Q+SBSkQOVLLU8gBADcU0FAShV5BXgm35
kBGl+/D1MBJLt6q8GZWHMWIHOX4GN28A/PEemaKg3dZHEtPM3w==
=su9m
-----END PGP SIGNATURE-----"""


@pytest.fixture
def secret_message():
    return """\
-----BEGIN PGP MESSAGE-----

hIwDVTqCoFjAx5UBA/wIt5OUfsKV2VPB2P+c6r7xVvPIPiA1FjNpU2x1G8A/dxVq
kAOhXJ9KkM6yon0PJReF3w8QPgZCo5tCmwqMtin4OY/WTw1ExyIWIaS7XJh1ktPM
TJL7RpyeywGHiAveLs9rznZtVwi0xg+rTSWpoMS/8GbKpOyf3twWMsiFfndr09JJ
ASWYXtfsUT3IVA5dP0Mr3/Yg0v90d+X2RqUHM+sUiUtwh4mb+vUcm7UOQRyGAR4V
h7jTNclSQwWCGzx6OaWKnrCVafRXbH4aeA==
=tw4x
-----END PGP MESSAGE-----"""


@pytest.fixture(params=["a"])
def sig(request, tmp_path):
    sigs = "\n".join(request.getfixturevalue(f"key_{x}_sig") for x in request.param)
    sig = tmp_path / "sig.asc"
    sig.write_text(sigs)
    yield str(sig)
    sig.unlink()


@pytest.fixture
def gnupg(gpghome):
    return gnupglib.GPG(gnupghome=str(gpghome))


@pytest.fixture
def gnupg_keyring(gpghome, keyring):
    return gnupglib.GPG(gnupghome=str(gpghome), keyring=keyring)


@pytest.fixture
def gnupg_privkeyring(gpghome, keyring_privkeys):
    return gnupglib.GPG(gnupghome=str(gpghome), keyring=keyring_privkeys)


@pytest.fixture(params=["abcde"])
def _pubkeys_present(gnupg, request):
    pubkeys = [request.getfixturevalue(f"key_{x}_pub") for x in request.param]
    fingerprints = [request.getfixturevalue(f"key_{x}_fp") for x in request.param]
    gnupg.import_keys("\n".join(pubkeys))
    present_keys = gnupg.list_keys()
    for fp in fingerprints:
        assert any(x["fingerprint"] == fp for x in present_keys)
    yield
    # cleanup is taken care of by gpghome and tmp_path


@pytest.fixture(params=["ab"])
def _privkeys_present(gnupg, request):
    privkeys = [request.getfixturevalue(f"key_{x}_priv") for x in request.param]
    fingerprints = [request.getfixturevalue(f"key_{x}_fp") for x in request.param]
    res = gnupg.import_keys("\n".join(privkeys))
    assert set(res.fingerprints) == set(fingerprints)
    present_keys = gnupg.list_keys(secret=True)
    assert present_keys
    for fp in fingerprints:
        assert any(x["fingerprint"] == fp for x in present_keys)
    yield
    # cleanup is taken care of by gpghome and tmp_path


@pytest.fixture(params=["a"])
def keyring(gpghome, tmp_path, request):
    keyring = tmp_path / "keys.gpg"
    _gnupg_keyring = gnupglib.GPG(gnupghome=str(gpghome), keyring=str(keyring))
    pubkeys = [request.getfixturevalue(f"key_{x}_pub") for x in request.param]
    fingerprints = [request.getfixturevalue(f"key_{x}_fp") for x in request.param]
    _gnupg_keyring.import_keys("\n".join(pubkeys))
    present_keys = _gnupg_keyring.list_keys()
    for fp in fingerprints:
        assert any(x["fingerprint"] == fp for x in present_keys)
    yield str(keyring)
    # cleanup is taken care of by gpghome and tmp_path


@pytest.fixture(params=["a"])
def keyring_privkeys(gpghome, gnupg, tmp_path, request):
    keyring = tmp_path / "keys.gpg"
    _gnupg_keyring = gnupglib.GPG(gnupghome=str(gpghome), keyring=str(keyring))
    privkeys = [request.getfixturevalue(f"key_{x}_priv") for x in request.param]
    fingerprints = [request.getfixturevalue(f"key_{x}_fp") for x in request.param]
    _gnupg_keyring.import_keys("\n".join(privkeys))
    present_privkeys = _gnupg_keyring.list_keys(secret=True)
    assert present_privkeys
    for fp in fingerprints:
        assert any(x["fingerprint"] == fp for x in present_privkeys)
    yield str(keyring)
    # cleanup is taken care of by gpghome and tmp_path


@pytest.mark.usefixtures("_pubkeys_present")
def test_list_keys(gpg, gpghome, gnupg):
    res = gpg.list_keys(gnupghome=str(gpghome))
    assert res
    assert len(res) == len(gnupg.list_keys())


def test_list_keys_in_keyring(gpg, gpghome, keyring, gnupg_keyring):
    res = gpg.list_keys(gnupghome=str(gpghome), keyring=keyring)
    assert len(res) == len(gnupg_keyring.list_keys())


@pytest.mark.usefixtures("_privkeys_present")
def test_list_secret_keys(gpghome, gpg, gnupg):
    res = gpg.list_secret_keys(gnupghome=str(gpghome))
    assert len(res) == len(gnupg.list_keys(secret=True))


def test_list_secret_keys_in_keyring(gpghome, gpg, keyring_privkeys, gnupg_privkeyring):
    res = gpg.list_secret_keys(gnupghome=str(gpghome), keyring=keyring_privkeys)
    assert len(res) == len(gnupg_privkeyring.list_keys(secret=True))


@pytest.mark.requires_random_entropy
def test_create_key(gpghome, gpg, gnupg):
    res = gpg.create_key(gnupghome=str(gpghome))
    assert res
    assert "message" in res
    assert "successfully generated" in res["message"]
    assert "fingerprint" in res
    assert res["fingerprint"]
    assert gnupg.list_keys(secret=True, keys=res["fingerprint"])


@pytest.mark.requires_random_entropy
def test_create_key_in_keyring(gpghome, gpg, gnupg, keyring, gnupg_keyring):
    res = gpg.create_key(gnupghome=str(gpghome), keyring=keyring)
    assert res
    assert "message" in res
    assert "successfully generated" in res["message"]
    assert "fingerprint" in res
    assert res["fingerprint"]
    assert not gnupg.list_keys(secret=True, keys=res["fingerprint"])
    assert gnupg_keyring.list_keys(secret=True, keys=res["fingerprint"])


@pytest.mark.usefixtures("_pubkeys_present")
def test_delete_key(gpghome, gpg, gnupg, key_a_fp):
    assert gnupg.list_keys(keys=key_a_fp)
    res = gpg.delete_key(
        fingerprint=key_a_fp, gnupghome=str(gpghome), use_passphrase=False
    )
    assert res["res"]
    assert not gnupg.list_keys(keys=key_a_fp)


@pytest.mark.usefixtures("_pubkeys_present")
def test_delete_key_from_keyring(gpghome, gpg, key_a_fp, keyring, gnupg, gnupg_keyring):
    assert gnupg.list_keys(keys=key_a_fp)
    assert gnupg_keyring.list_keys(keys=key_a_fp)
    res = gpg.delete_key(
        fingerprint=key_a_fp,
        gnupghome=str(gpghome),
        keyring=keyring,
        use_passphrase=False,
    )
    assert res["res"]
    assert gnupg.list_keys(keys=key_a_fp)
    assert not gnupg_keyring.list_keys(keys=key_a_fp)


@pytest.mark.usefixtures("_pubkeys_present")
def test_get_key(gpghome, gpg, key_a_fp):
    res = gpg.get_key(fingerprint=key_a_fp, gnupghome=str(gpghome))
    assert res
    assert "keyid" in res
    assert res["keyid"] == key_a_fp[-16:]
    assert "keyLength" in res
    assert res["keyLength"] == "1024"


def test_get_key_from_keyring(gpghome, gpg, key_a_fp, keyring, gnupg):
    assert not gnupg.list_keys()
    res = gpg.get_key(fingerprint=key_a_fp, gnupghome=str(gpghome), keyring=keyring)
    assert res
    assert "keyid" in res
    assert res["keyid"] == key_a_fp[-16:]
    assert "keyLength" in res
    assert res["keyLength"] == "1024"


@pytest.mark.usefixtures("_privkeys_present")
def test_get_secret_key(gpghome, gpg, key_a_fp):
    res = gpg.get_secret_key(fingerprint=key_a_fp, gnupghome=str(gpghome))
    assert res
    assert "keyid" in res
    assert res["keyid"] == key_a_fp[-16:]
    assert "keyLength" in res
    assert res["keyLength"] == "1024"


def test_get_secret_key_from_keyring(gpghome, gpg, key_a_fp, keyring_privkeys, gnupg):
    assert not gnupg.list_keys(keys=key_a_fp, secret=True)
    res = gpg.get_secret_key(
        fingerprint=key_a_fp, gnupghome=str(gpghome), keyring=keyring_privkeys
    )
    assert res
    assert "keyid" in res
    assert res["keyid"] == key_a_fp[-16:]
    assert "keyLength" in res
    assert res["keyLength"] == "1024"


def test_import_key(gpghome, gnupg, gpg, key_a_pub, key_a_fp):
    assert not gnupg.list_keys(keys=key_a_fp)
    res = gpg.import_key(text=key_a_pub, gnupghome=str(gpghome))
    assert res
    assert res["res"]
    assert "Successfully imported" in res["message"]
    assert gnupg.list_keys(keys=key_a_fp)


@pytest.mark.parametrize("keyring", [""], indirect=True)
def test_import_key_to_keyring(
    gpghome, gnupg, gpg, key_d_pub, key_d_fp, keyring, gnupg_keyring
):
    assert not gnupg.list_keys(keys=key_d_fp)
    assert not gnupg_keyring.list_keys(keys=key_d_fp)
    res = gpg.import_key(text=key_d_pub, gnupghome=str(gpghome), keyring=keyring)
    assert res
    assert res["res"]
    assert "Successfully imported" in res["message"]
    assert not gnupg.list_keys(keys=key_d_fp)
    assert gnupg_keyring.list_keys(keys=key_d_fp)


@pytest.mark.usefixtures("_pubkeys_present")
def test_export_key(gpghome, gpg, key_a_fp):
    res = gpg.export_key(keyids=key_a_fp, gnupghome=str(gpghome))
    assert res["res"]
    assert res["comment"].startswith("-----BEGIN PGP PUBLIC KEY BLOCK-----")
    assert res["comment"].endswith("-----END PGP PUBLIC KEY BLOCK-----\n")


def test_export_key_from_keyring(gpghome, gnupg, gpg, key_a_fp, keyring, gnupg_keyring):
    assert not gnupg.list_keys(keys=key_a_fp)
    assert gnupg_keyring.list_keys(keys=key_a_fp)
    res = gpg.export_key(keyids=key_a_fp, gnupghome=str(gpghome), keyring=keyring)
    assert res["res"]
    assert res["comment"].startswith("-----BEGIN PGP PUBLIC KEY BLOCK-----")
    assert res["comment"].endswith("-----END PGP PUBLIC KEY BLOCK-----\n")


@pytest.mark.usefixtures("_pubkeys_present")
@pytest.mark.parametrize("use_keyid", [True, False])
def test_trust_key(gpghome, key_a_fp, gnupg, gpg, use_keyid):
    keyid = key_a_fp[-16:] if use_keyid else None
    fingerprint = None if keyid else key_a_fp
    res = gpg.trust_key(
        fingerprint=fingerprint,
        keyid=keyid,
        trust_level="ultimately",
        gnupghome=str(gpghome),
    )
    assert res["res"]
    assert "fingerprint" in res
    assert res["fingerprint"] == key_a_fp
    key_info = gnupg.list_keys(keys=key_a_fp)
    assert key_info
    assert key_info[0]["trust"] == "u"


@pytest.mark.parametrize("use_keyid", [True, False])
def test_trust_key_keyring(
    gpghome, key_a_fp, keyring, gnupg, gnupg_keyring, gpg, use_keyid
):
    assert not gnupg.list_keys(keys=key_a_fp)
    keyid = key_a_fp[-16:] if use_keyid else None
    fingerprint = None if keyid else key_a_fp
    res = gpg.trust_key(
        fingerprint=fingerprint,
        keyid=keyid,
        trust_level="ultimately",
        gnupghome=str(gpghome),
        keyring=keyring,
    )
    assert res["res"]
    assert "fingerprint" in res
    assert res["fingerprint"] == key_a_fp
    key_info = gnupg_keyring.list_keys(keys=key_a_fp)
    assert key_info
    assert key_info[0]["trust"] == "u"


@pytest.mark.usefixtures("_privkeys_present")
@pytest.mark.requires_random_entropy
def test_sign(gpghome, gpg, gnupg, key_a_fp):
    assert gnupg.list_keys(secret=True, keys=key_a_fp)
    res = gpg.sign(text="foo", keyid=key_a_fp, gnupghome=str(gpghome))
    assert res
    assert res.startswith(b"-----BEGIN PGP SIGNED MESSAGE-----")
    assert res.endswith(b"-----END PGP SIGNATURE-----\n")


@pytest.mark.requires_random_entropy
def test_sign_with_keyring(
    gpghome, gpg, gnupg, key_a_fp, gnupg_privkeyring, keyring_privkeys
):
    assert not gnupg.list_keys(keys=key_a_fp, secret=True)
    assert gnupg_privkeyring.list_keys(keys=key_a_fp, secret=True)
    res = gpg.sign(
        text="foo", keyid=key_a_fp, gnupghome=str(gpghome), keyring=keyring_privkeys
    )
    assert res
    assert res.startswith(b"-----BEGIN PGP SIGNED MESSAGE-----")
    assert res.endswith(b"-----END PGP SIGNATURE-----\n")


@pytest.mark.parametrize(
    "sig,expected",
    [
        (["a"], True),
        (["c_fail"], False),
    ],
    indirect=["sig"],
)
@pytest.mark.usefixtures("_pubkeys_present")
def test_gpg_verify(gpg, gpghome, signed_data, sig, expected, key_a_fp):
    res = gpg.verify(filename=str(signed_data), gnupghome=str(gpghome), signature=sig)
    assert res["res"] is expected
    if not expected:
        assert "could not be verified" in res["message"]
    else:
        assert "is verified" in res["message"]
        assert "key_id" in res
        assert res["key_id"] == key_a_fp[-16:]


@pytest.mark.parametrize(
    "sig,by,expected",
    [
        ("a", "ab", True),
        ("bd", "ab", True),
        ("bd", "ad", True),
        ("a", "bcde", False),
        (["c_fail"], "bc", False),
        (["c_fail", "b"], "bc", True),
        # Bad signatures partly overwrite the preceding good one currently.
        # https://github.com/vsajip/python-gnupg/issues/214
        # The module contains a workaround for this issue though.
        # The following would not have an issue since we only rely on trust_level and fingerprint.
        (["b", "c_fail"], "bc", True),
        (["e_exp"], "e", False),
        (["e_exp", "a"], "a", True),
        (["a", "e_exp"], "a", True),
        ("f", "abcde", False),
        ("fa", "abcde", True),
        # The above mentioned issue also affects signatures
        # whose pubkeys are absent from the keychain, but they
        # also overwrite the previous one's fingerprint (which we compare to).
        # So the following would fail without the workaround:
        ("af", "abcde", True),
        # Without the workaround, the following would be accepted (= test failure),
        # even though the `f` signature was never verified because the pubkey
        # was missing.
        ("af", "f", False),
    ],
    indirect=["sig"],
)
@pytest.mark.usefixtures("_pubkeys_present")
def test_gpg_verify_signed_by_any(
    gpg, gpghome, signed_data, sig, by, expected, request
):
    fps = [request.getfixturevalue(f"key_{x}_fp") for x in by]
    res = gpg.verify(
        filename=str(signed_data),
        gnupghome=str(gpghome),
        signature=sig,
        signed_by_any=fps,
    )
    assert res["res"] is expected


@pytest.mark.parametrize(
    "sig,by,expected",
    [
        ("a", "a", True),
        ("bd", "a", False),
        ("ad", "ad", True),
        ("ad", "ad", True),
        ("ad", "ac", False),
        (["a", "c_fail"], "ac", False),
        # Bad signatures partly overwrite the preceding good one currently.
        # https://github.com/vsajip/python-gnupg/issues/214
        # The module contains a workaround for this issue though.
        # The following would not have an issue since we only rely on trust_level and fingerprint.
        (["a", "b", "c_fail"], "ab", True),
        (["a", "b", "c_fail"], "abc", False),
        (["c_fail", "a", "b"], "ab", True),
        ("fad", "da", True),
        ("fd", "da", False),
        # The above mentioned issue also affects signatures
        # whose pubkeys are absent from the keychain, but they
        # also overwrite the previous one's fingerprint (which we compare to).
        # So the following would fail without the workaround:
        ("dfa", "da", True),
        # Without the workaround, the following would be accepted (= test fail),
        # even though the `f` signature was never verified because the pubkey
        # was missing.
        ("abf", "af", False),
    ],
    indirect=["sig"],
)
@pytest.mark.usefixtures("_pubkeys_present")
def test_gpg_verify_signed_by_all(
    gpg, gpghome, signed_data, sig, by, expected, request
):
    fps = [request.getfixturevalue(f"key_{x}_fp") for x in by]
    res = gpg.verify(
        filename=str(signed_data),
        gnupghome=str(gpghome),
        signature=sig,
        signed_by_all=fps,
    )
    assert res["res"] is expected


@pytest.mark.usefixtures("_pubkeys_present")
def test_verify(gpghome, gpg, sig, signed_data, key_a_fp):
    res = gpg.verify(
        filename=str(signed_data),
        signature=str(sig),
        gnupghome=str(gpghome),
    )
    assert res["res"]
    assert "is verified" in res["message"]
    assert "key_id" in res
    assert res["key_id"] == key_a_fp[-16:]


def test_verify_with_keyring(gpghome, gnupg, gpg, keyring, sig, signed_data, key_a_fp):
    assert not gnupg.list_keys(keys=key_a_fp)
    res = gpg.verify(
        filename=str(signed_data),
        signature=str(sig),
        gnupghome=str(gpghome),
        keyring=keyring,
    )
    assert res["res"]
    assert "is verified" in res["message"]
    assert "key_id" in res
    assert res["key_id"] == key_a_fp[-16:]


@pytest.mark.usefixtures("_pubkeys_present")
# Can't easily test the other signature validity levels since
# we would need to sign the pubkey ourselves, which is not
# exposed by python-gnupg as of release 0.5.2.
@pytest.mark.parametrize(
    "ownertrust,text", (("TRUST_NEVER", "Undefined"), ("TRUST_ULTIMATE", "Ultimate"))
)
def test_verify_trust_levels(
    gpghome, gpg, gnupg, key_a_fp, sig, signed_data, ownertrust, text
):
    gnupg.trust_keys(key_a_fp, ownertrust)
    res = gpg.verify(
        filename=str(signed_data),
        signature=sig,
        gnupghome=str(gpghome),
    )
    assert res["res"] is True
    assert "is verified" in res["message"]
    assert "key_id" in res
    assert res["key_id"] == key_a_fp[-16:]
    assert res["trust_level"] == text


@pytest.mark.usefixtures("_pubkeys_present")
@pytest.mark.requires_random_entropy
def test_encrypt(gpghome, gpg, gnupg, key_b_fp):
    assert gnupg.list_keys(keys=key_b_fp)
    res = gpg.encrypt(
        text="I like turtles",
        recipients=key_b_fp,
        gnupghome=str(gpghome),
        always_trust=True,
    )
    assert res
    assert res["res"]
    assert res["comment"]
    assert res["comment"].startswith(b"-----BEGIN PGP MESSAGE-----")
    assert res["comment"].endswith(b"-----END PGP MESSAGE-----\n")


@pytest.mark.requires_random_entropy
def test_encrypt_with_keyring(gpghome, gpg, gnupg, key_a_fp, keyring, gnupg_keyring):
    assert not gnupg.list_keys(keys=key_a_fp)
    assert gnupg_keyring.list_keys(keys=key_a_fp)
    res = gpg.encrypt(
        text="I like turtles",
        recipients=key_a_fp,
        gnupghome=str(gpghome),
        keyring=keyring,
        always_trust=True,
    )
    assert res
    assert res["res"]
    assert res["comment"]
    assert res["comment"].startswith(b"-----BEGIN PGP MESSAGE-----")
    assert res["comment"].endswith(b"-----END PGP MESSAGE-----\n")


@pytest.mark.usefixtures("_privkeys_present")
def test_decrypt(gpghome, gpg, gnupg, secret_message, key_a_fp):
    assert gnupg.list_keys(secret=True, keys=key_a_fp)
    res = gpg.decrypt(text=secret_message, gnupghome=str(gpghome))
    assert res["res"]
    assert res["comment"]
    assert res["comment"] == b"I like turtles"


def test_decrypt_with_keyring(
    gpghome, gpg, gnupg, gnupg_privkeyring, keyring_privkeys, secret_message, key_a_fp
):
    assert not gnupg.list_keys(secret=True, keys=key_a_fp)
    assert gnupg_privkeyring.list_keys(secret=True, keys=key_a_fp)
    res = gpg.decrypt(
        text=secret_message, gnupghome=str(gpghome), keyring=keyring_privkeys
    )
    assert res["res"]
    assert res["comment"]
    assert res["comment"] == b"I like turtles"
